#	This file is part of SurrealServices.
#
#	SurrealServices is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	SurrealServices is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with SurrealServices; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package nickserv;

#require Exporter;
#@ISA = qw(Exporter);
#@EXPORT_OK = qw(get_user_id get_user_nick is_identified notice);

use strict;
use threads::shared;
use Time::Local;
use DBI qw(:sql_types);
#use MIME::Base64;

use constant {
	MAX_LIM => 16777215,

	#current nickreg.flags definition limits us to 16 of these. or 32768 as last flag
	F_HIDEMAIL => 1,
	F_NOMEMO => 2,
	F_NOACC => 4,
	F_NEVEROP => 8,
	F_AUTH => 16,
	F_HOLD => 32,
	F_FREEZE => 64,
	F_VACATION => 128,
	F_EMAILREG => 256,

	UF_FINISHED => 1,

	OLD_USER_AGE => 300,

	NTF_QUIT => 1,
	NTF_GREET => 2,
	# 3 will be NTF_JOIN
	NTF_AUTH => 4,
	NTF_UMODE => 5,
	NTF_VACATION => 6,
	NTF_EMAILREG => 7,
};

*synced = \&net::synced;

our $user_table : shared;
my %enforcers : shared;
my %preenforce : shared;
our %uidcache : shared;
my %newuser : shared;
my %olduser : shared;

our $nsnick = 'NickServ';

my $cur_lock;
my $cnt_lock = 0;

my @protect_short = ('none', 'normal', 'high', 'kill');
my @protect_long = (
	'You will not be required to identify to use this nick.',
	'You must identify within 60 seconds to use this nick.',
	'You must identify before using this nick.',
	'You must identify before using this nick or you will be disconnected.'
);
our %protect_level = (
	'none'		=> 0,
	'no'		=> 0,
	'false'		=> 0,
	'off'		=> 0,
	'0'		=> 0,

	'true'		=> 1,
	'yes'		=> 1,
	'on'		=> 1,
	'normal'	=> 1,
	'1'		=> 1,

	'high'		=> 2,
	'2'		=> 2,

	'kill'		=> 3,
	'3'		=> 3
);
	
my $err_deny = 'Permission denied.';
my $err_email = 'Your email address looks funny.';
my $err_pass = 'Invalid password.';

our (
	$nick_check, $nick_create, $nick_create_old, $nick_change, $nick_quit, $nick_delete, $nick_id_delete, $get_quit_empty_chans, $nick_chan_delete,
	$get_hostless_nicks,

	$get_squit_lock, $squit_users, $squit_nickreg, $get_squit_empty_chans,

	$get_nickchg, $del_nickchg_id, $add_nickchg, $reap_nickchg,
	
	$get_user_id, $get_user_nick,
	$get_nick_inval, $inc_nick_inval,
	$get_id_nicks, $get_using_nicks, $is_identified, $is_registered,
	$is_alias_of,

	$get_guest, $set_guest,

	$get_lock, $release_lock,
	
	$get_umodes, $set_umodes,
	
	$get_info, $get_host, $get_vhost, $set_vhost, $set_ident, $set_ip, $get_user_info,
	$update_regnick_vhost, $get_regd_time, $get_nickreg_quit, 

	$chk_clone_except, $count_clones,
	
	$get_pass, $get_pass_email, 
	$set_pass, $set_email, $set_protect_level,
	
	$set_flags, $get_flags, $set_flag, $unset_flag,

	$get_user_flags, $set_user_flag, $unset_user_flag, $set_user_flag_all,

	$get_root_nick, $chk_pass, $identify, $identify_ign, $id_update, $logout, $unidentify, $unidentify_single,
	$update_lastseen, $quit_update, $protect_level, $update_nickalias_last,

	$get_register_lock, $register, $create_alias, $drop, $change_root,

	$get_aliases, $get_glist, $count_aliases, $get_random_alias, $delete_alias, $delete_aliases,
	$get_all_access, $del_all_access, $change_all_access, $change_akicks, $change_founders,
	$change_successors, $change_svsops,
	
	$lock_user_table, $cleanup, $unlock_tables,
	$is_online,

	$get_matching_nicks,

	$cleanup_nickid, $cleanup_users, $get_expired, $get_near_expired, $set_near_expired,
	
	$get_watches, $check_watch, $set_watch, $del_watch, $drop_watch,
	$get_silences, $check_silence, $set_silence, $del_silence, $drop_silence, $get_expired_silences,

	$get_seen,

	$set_greet, $get_greet, $del_greet,
	$get_num_nicktext_type, $drop_nicktext,

	$get_auth_chan, $get_auth_num, $del_auth, $list_auth, $add_auth,

	$set_umode_ntf, $del_umode_ntf, $get_umode_ntf,

	$set_vacation_ntf, $get_vacation_ntf,

	$set_emailreg_ntf, $get_emailreg_ntf, $del_emailreg_ntf,
);

sub init() {
	$nick_check = $services::dbh->prepare("SELECT id FROM user WHERE nick=? AND online=0 AND time=?");
	$nick_create = $services::dbh->prepare("INSERT INTO user SET id=(RAND()*294967293)+1, nick=?, time=?, inval=0, ident=?, host=?, vhost=?, server=?, modes=?, gecos=?, flags=?, online=1");
	$nick_create_old = $services::dbh->prepare("UPDATE user SET ident=?, host=?, vhost=?, server=?, modes=?, gecos=?, flags=?, online=1 WHERE nick=?");
	$nick_change = $services::dbh->prepare("UPDATE user SET nick=?, time=? WHERE nick=?");
	$nick_quit = $services::dbh->prepare("UPDATE user SET online=0, quittime=UNIX_TIMESTAMP() WHERE nick=?");
	$nick_delete = $services::dbh->prepare("DELETE FROM user WHERE nick=?");
	$nick_id_delete = $services::dbh->prepare("DELETE FROM nickid WHERE id=?");
	$get_quit_empty_chans = $services::dbh->prepare("SELECT cu2.chan, COUNT(*) AS c
		FROM chanuser AS cu1, chanuser AS cu2
		WHERE cu1.nickid=?
		AND cu1.chan=cu2.chan AND cu1.joined=1 AND cu2.joined=1
		GROUP BY cu2.chan HAVING c=1 ORDER BY NULL");
	$nick_chan_delete = $services::dbh->prepare("DELETE FROM chanuser WHERE nickid=?");
	$get_hostless_nicks = $services::dbh->prepare("SELECT nick FROM user WHERE vhost='*'");

	$get_squit_lock = $services::dbh->prepare("LOCK TABLES chanuser WRITE, chanuser AS cu1 READ LOCAL, chanuser AS cu2 READ LOCAL, user WRITE, nickreg WRITE, nickid WRITE, chanban WRITE, chan WRITE, chanreg READ LOCAL");
	$squit_users = $services::dbh->prepare("UPDATE chanuser, user
		SET chanuser.joined=0, user.online=0, user.quittime=UNIX_TIMESTAMP()
		WHERE user.id=chanuser.nickid AND user.server=?");
	# Must call squit_nickreg first as squit_users modifies user.flags
	$squit_nickreg = $services::dbh->prepare("UPDATE nickreg, nickid, user
		SET nickreg.quit=?, nickreg.last=UNIX_TIMESTAMP()
		WHERE nickreg.nick=nickid.nick AND nickid.id=user.id
		AND user.online=1 AND user.server=?");
	$get_squit_empty_chans = $services::dbh->prepare("SELECT cu2.chan, COUNT(*) AS c
		FROM user, chanuser AS cu1, chanuser AS cu2
		WHERE user.server=? AND cu1.nickid=user.id
		AND cu1.chan=cu2.chan AND cu1.joined=1 AND cu2.joined=1
		GROUP BY cu2.chan HAVING c=1 ORDER BY NULL");

	$get_nickchg = $services::dbh->prepare("SELECT nickid FROM nickchg WHERE nick=?");
	$del_nickchg_id = $services::dbh->prepare("DELETE FROM nickchg WHERE nickid=?");
	$add_nickchg = $services::dbh->prepare("REPLACE INTO nickchg SELECT ?, id, ? FROM user WHERE nick=?");
	$reap_nickchg = $services::dbh->prepare("DELETE FROM nickchg WHERE seq<?");

	$get_user_id = $services::dbh->prepare("SELECT id FROM user WHERE nick=?");
	$get_user_nick = $services::dbh->prepare("SELECT nick FROM user WHERE id=?");
	
	$get_nick_inval = $services::dbh->prepare("SELECT nick, inval FROM user WHERE id=?");
	$inc_nick_inval = $services::dbh->prepare("UPDATE user SET inval=inval+1 WHERE id=?");

	$get_id_nicks = $services::dbh->prepare("SELECT nickreg.nick FROM nickid, nickreg WHERE nickid.nrid=nickreg.id AND nickid.id=?");
	$get_using_nicks = $services::dbh->prepare("SELECT user.nick, user.id FROM user, nickid, nickalias WHERE user.id=nickid.id AND nickid.nrid=nickalias.nrid AND nickalias.alias=? AND user.online=1");
	$is_identified = $services::dbh->prepare("SELECT 1 FROM user, nickid, nickalias WHERE user.id=nickid.id AND user.nick=? AND nickid.nrid=nickalias.nrid AND nickalias.alias=?");
	$is_registered = $services::dbh->prepare("SELECT 1 FROM nickalias WHERE alias=?");
	$is_alias_of = $services::dbh->prepare("SELECT 1 FROM nickalias AS n1 LEFT JOIN nickalias AS n2 ON n1.nrid=n2.nrid WHERE n1.alias=? AND n2.alias=? LIMIT 1");

	$get_guest = $services::dbh->prepare("SELECT guest FROM user WHERE nick=?");
	$set_guest = $services::dbh->prepare("UPDATE user SET guest=? WHERE nick=?");

	$get_lock = $services::dbh->prepare("SELECT GET_LOCK(?, 10)");
	$release_lock = $services::dbh->prepare("SELECT RELEASE_LOCK(?)");

	$get_umodes = $services::dbh->prepare("SELECT modes FROM user WHERE id=?");
	$set_umodes = $services::dbh->prepare("UPDATE user SET modes=? WHERE id=?");
	
	$get_info = $services::dbh->prepare("SELECT nickreg.email, nickreg.regd, nickreg.last, nickreg.flags, nickreg.ident,
		nickreg.vhost, nickreg.gecos, nickalias.last
		FROM nickreg, nickalias WHERE nickalias.nrid=nickreg.id AND nickalias.alias=?");
	$get_nickreg_quit = $services::dbh->prepare("SELECT nicktext.data FROM nickreg, nicktext, nickalias
		WHERE nickalias.nrid=nickreg.id AND nickalias.alias=? AND
		(nicktext.nrid=nickreg.id AND nicktext.type=".NTF_QUIT.")");
	$get_host = $services::dbh->prepare("SELECT ident, host FROM user WHERE id=?");
	$get_vhost = $services::dbh->prepare("SELECT ident, vhost FROM user WHERE id=?");
	$set_ident = $services::dbh->prepare("UPDATE user SET ident=? WHERE id=?");
	$set_vhost = $services::dbh->prepare("UPDATE user SET vhost=? WHERE id=?");
	$set_ip = $services::dbh->prepare("UPDATE user SET ip=? WHERE id=?");
	$get_user_info  = $services::dbh->prepare("SELECT ident, host, vhost, gecos, server FROM user WHERE id=?");
	$update_regnick_vhost = $services::dbh->prepare("UPDATE nickreg,nickid SET nickreg.vhost=?
		WHERE nickreg.id=nickid.nrid AND nickid.id=?");
	$get_regd_time = $services::dbh->prepare("SELECT nickreg.regd FROM nickreg, nickalias
		WHERE nickalias.nrid=nickreg.id and nickalias.alias=?");

	$chk_clone_except = $services::dbh->prepare("SELECT GREATEST(IF((user.ip >> (32 - sesexip.mask)) = (sesexip.ip >> (32 - sesexip.mask)), sesexip.lim, 0), IF(IF(sesexname.serv, user.server, user.host) LIKE sesexname.host, sesexname.lim, 0)) AS n FROM user, sesexip, sesexname WHERE user.id=? ORDER BY n DESC LIMIT 1");
	$count_clones = $services::dbh->prepare("SELECT COUNT(*) FROM user WHERE ip=? AND online=1");

	$get_root_nick = $services::dbh->prepare("SELECT nickreg.nick FROM nickreg, nickalias WHERE nickreg.id=nickalias.nrid AND nickalias.alias=?");
	$chk_pass = $services::dbh->prepare("SELECT 1 FROM nickreg WHERE nick=? AND pass=?");
	$identify = $services::dbh->prepare("INSERT INTO nickid SELECT ?, id FROM nickreg WHERE nick=?");
	$identify_ign = $services::dbh->prepare("INSERT IGNORE INTO nickid SELECT ?, id FROM nickreg WHERE nick=?");
	$id_update = $services::dbh->prepare("UPDATE nickreg, user SET
		nickreg.last=UNIX_TIMESTAMP(), nickreg.ident=user.ident,
		nickreg.vhost=user.vhost, nickreg.gecos=user.gecos,
		nickreg.nearexp=0, nickreg.flags = (nickreg.flags & ~". F_VACATION .")
		WHERE nickreg.nick=? AND user.id=?");
	$logout = $services::dbh->prepare("DELETE FROM nickid WHERE id=?");
	$unidentify = $services::dbh->prepare("DELETE FROM nickid USING nickreg, nickid WHERE nickreg.nick=? AND nickid.nrid=nickreg.id");

	$update_lastseen = $services::dbh->prepare("UPDATE nickreg,nickid SET nickreg.last=UNIX_TIMESTAMP()
		WHERE nickreg.id=nickid.nrid AND nickid.id=?");
	$update_nickalias_last = $services::dbh->prepare("UPDATE nickalias SET last=UNIX_TIMESTAMP() WHERE alias=?");
	$quit_update = $services::dbh->prepare("REPLACE INTO nicktext
		SELECT nickreg.id, ".NTF_QUIT().", 0, NULL, ? FROM nickreg, nickid
		WHERE nickreg.id=nickid.nrid AND nickid.id=?");
	$protect_level = $services::dbh->prepare("SELECT protect FROM nickalias WHERE alias=?");

	$set_protect_level = $services::dbh->prepare("UPDATE nickalias SET protect=? WHERE alias=?");
	$set_email = $services::dbh->prepare("UPDATE nickreg, nickalias SET nickreg.email=? WHERE nickalias.nrid=nickreg.id AND nickalias.alias=?");
	$set_pass = $services::dbh->prepare("UPDATE nickreg, nickalias SET nickreg.pass=? WHERE nickalias.nrid=nickreg.id AND nickalias.alias=?");
	
	$set_flags = $services::dbh->prepare("UPDATE nickreg, nickalias SET nickreg.flags=? WHERE nickalias.nrid=nickreg.id AND nickalias.alias=?");
	$get_flags = $services::dbh->prepare("SELECT nickreg.flags FROM nickreg, nickalias WHERE nickalias.nrid=nickreg.id AND nickalias.alias=?");
	$set_flag = $services::dbh->prepare("UPDATE nickreg, nickalias SET nickreg.flags=IF(nickreg.flags & ?, nickreg.flags, nickreg.flags ^ ?) WHERE nickalias.nrid=nickreg.id AND nickalias.alias=?");
	$unset_flag = $services::dbh->prepare("UPDATE nickreg, nickalias SET nickreg.flags=IF(nickreg.flags & ?, nickreg.flags ^ ?, nickreg.flags) WHERE nickalias.nrid=nickreg.id AND nickalias.alias=?");

	$get_user_flags = $services::dbh->prepare("SELECT flags FROM user WHERE id=?");
	$set_user_flag = $services::dbh->prepare("UPDATE user SET flags=IF(flags & ?, flags, flags ^ ?) WHERE id=?");
	$unset_user_flag = $services::dbh->prepare("UPDATE user SET flags=IF(flags & ?, flags ^ ?, flags) WHERE id=?");
	$set_user_flag_all = $services::dbh->prepare("UPDATE user SET flags=flags | ?");

	$get_pass = $services::dbh->prepare("SELECT nickreg.pass FROM nickreg, nickalias WHERE nickalias.nrid=nickreg.id AND nickalias.alias=?");
	$get_pass_email  = $services::dbh->prepare("SELECT nickreg.pass, nickreg.email FROM nickalias, nickreg WHERE nickreg.id=nickalias.nrid AND nickalias.alias=?");

	$get_register_lock = $services::dbh->prepare("LOCK TABLES nickalias WRITE, nickreg WRITE");
	$register = $services::dbh->prepare("INSERT INTO nickreg SET nick=?, pass=?, email=?, flags=".F_HIDEMAIL().", regd=UNIX_TIMESTAMP(), last=UNIX_TIMESTAMP()");
	$create_alias = $services::dbh->prepare("INSERT INTO nickalias SELECT id, ?, NULL, NULL FROM nickreg WHERE nick=?");

	$drop = $services::dbh->prepare("DELETE FROM nickreg WHERE nick=?");
	
	$get_aliases = $services::dbh->prepare("SELECT nickalias.alias FROM nickalias, nickreg WHERE
		nickalias.nrid=nickreg.id AND nickreg.nick=? ORDER BY nickalias.alias");
	$get_glist = $services::dbh->prepare("SELECT nickalias.alias, nickalias.protect, nickalias.last 
		FROM nickalias, nickreg WHERE
		nickalias.nrid=nickreg.id AND nickreg.nick=? ORDER BY nickalias.alias");
	$count_aliases = $services::dbh->prepare("SELECT COUNT(*) FROM nickalias, nickreg WHERE
		nickalias.nrid=nickreg.id AND nickreg.nick=?");
	$get_random_alias = $services::dbh->prepare("SELECT nickalias.alias FROM nickalias, nickreg WHERE
		nickalias.nrid=nickreg.id AND nickreg.nick=? AND nickalias.alias != nickreg.nick LIMIT 1");
	$delete_alias = $services::dbh->prepare("DELETE FROM nickalias WHERE alias=?");
	$delete_aliases = $services::dbh->prepare("DELETE FROM nickalias USING nickreg, nickalias WHERE
		nickalias.nrid=nickreg.id AND nickreg.nick=?");
	
	$get_all_access = $services::dbh->prepare("SELECT chanacc.chan, chanacc.level, chanacc.adder, chanacc.time FROM nickalias, chanacc WHERE chanacc.nrid=nickalias.nrid AND nickalias.alias=? ORDER BY chanacc.chan");
	$del_all_access = $services::dbh->prepare("DELETE FROM chanacc USING chanacc, nickreg WHERE chanacc.nrid=nickreg.id AND nickreg.nick=?");
	
	$change_root = $services::dbh->prepare("UPDATE nickreg SET nick=? WHERE nick=?");

	$cleanup = $services::dbh->prepare("DELETE FROM nickid USING nickid LEFT JOIN user ON (nickid.id=user.id) WHERE user.id IS NULL");
	$unlock_tables = $services::dbh->prepare("UNLOCK TABLES");

	$is_online = $services::dbh->prepare("SELECT 1 FROM user WHERE nick=? AND online=1");

	$get_matching_nicks = $services::dbh->prepare("SELECT nickalias.alias, nickreg.nick, nickreg.ident, nickreg.vhost FROM nickalias, nickreg WHERE nickalias.nrid=nickreg.id AND nickalias.alias LIKE ? AND nickreg.ident LIKE ? AND nickreg.vhost LIKE ? LIMIT 50");
	
	$cleanup_nickid = $services::dbh->prepare("DELETE FROM nickid, user USING nickid LEFT JOIN user ON(nickid.id=user.id) WHERE user.id IS NULL OR (user.online=0 AND quittime<?)");
	$cleanup_users = $services::dbh->prepare("DELETE FROM user WHERE online=0 AND quittime<?");
	
	$get_expired = $services::dbh->prepare("SELECT nickreg.nick, nickreg.email, nickreg.ident, nickreg.vhost
		FROM nickreg LEFT JOIN nickid ON(nickreg.id=nickid.nrid)
		LEFT JOIN svsop ON(nickreg.id=svsop.nrid)
		WHERE nickid.nrid IS NULL AND svsop.nrid IS NULL ".
		($services::conf{'nearexpire'} ? 'AND nickreg.nearexp!=0 ' : '').
		"AND ( ( !(nickreg.flags & " . F_HOLD . ") AND !(nickreg.flags & " . F_VACATION . ") AND nickreg.last<? ) OR
		( (nickreg.flags & " . F_VACATION . ") AND nickreg.last<? ) OR
		( (nickreg.flags & ". F_EMAILREG .") AND nickreg.last<?)
		)");
	$get_near_expired = $services::dbh->prepare("SELECT nickreg.nick, nickreg.email
		FROM nickreg LEFT JOIN nickid ON(nickreg.id=nickid.nrid) 
		LEFT JOIN svsop ON(nickreg.id=svsop.nrid)
		WHERE nickid.nrid IS NULL AND svsop.nrid IS NULL AND nickreg.nearexp=0 AND
		( ( !(nickreg.flags & " . F_HOLD . ") AND !(nickreg.flags & " . F_VACATION . ") AND nickreg.last<? ) OR
		( (nickreg.flags & " . F_VACATION . ") AND nickreg.last<? )
		)");
	$set_near_expired = $services::dbh->prepare("UPDATE nickreg SET nearexp=1, last=IF(? > last+?, ?, last) WHERE nick=?");
	
	$get_watches = $services::dbh->prepare("SELECT watch.mask, watch.time FROM watch, nickreg WHERE watch.nrid=nickreg.id AND nickreg.nick=?");
	$check_watch = $services::dbh->prepare("SELECT 1 FROM watch, nickreg WHERE watch.nrid=nickreg.id AND nickreg.nick=? AND watch.mask=?");
	$set_watch = $services::dbh->prepare("INSERT INTO watch SELECT id, ?, ? FROM nickreg WHERE nick=?");
	$del_watch = $services::dbh->prepare("DELETE FROM watch USING watch, nickreg WHERE watch.nrid=nickreg.id AND nickreg.nick=? AND watch.mask=?");
	$drop_watch = $services::dbh->prepare("DELETE FROM watch USING nickreg, watch WHERE
		watch.nrid=nickreg.id AND nickreg.nick=?");

	$get_silences = $services::dbh->prepare("SELECT silence.mask, silence.time, silence.expiry, silence.comment FROM
		silence, nickreg WHERE silence.nrid=nickreg.id AND nickreg.nick=?");
	$check_silence = $services::dbh->prepare("SELECT 1 FROM silence, nickreg WHERE
		silence.nrid=nickreg.id AND nickreg.nick=? AND silence.mask=?");
	$set_silence = $services::dbh->prepare("INSERT INTO silence SELECT id, ?, ?, ?, ? FROM nickreg WHERE nick=?");
	$del_silence = $services::dbh->prepare("DELETE FROM silence USING silence, nickreg WHERE silence.nrid=nickreg.id AND
		nickreg.nick=? AND silence.mask=?");
	$drop_silence = $services::dbh->prepare("DELETE FROM silence USING nickreg, silence WHERE
		silence.nrid=nickreg.id AND nickreg.nick=?");
	$get_expired_silences = $services::dbh->prepare("SELECT nickreg.nick, silence.mask FROM nickreg, silence
		WHERE nickreg.id=silence.nrid AND silence.expiry < UNIX_TIMESTAMP() ORDER BY nickreg.nick");

	$get_seen = $services::dbh->prepare("SELECT nickalias.alias, nickreg.nick, nickreg.last FROM nickreg, nickalias 
		WHERE nickalias.nrid=nickreg.id AND nickalias.alias=?");

	$set_greet = $services::dbh->prepare("REPLACE INTO nicktext SELECT nickreg.id, ".NTF_GREET.", 0, NULL, ? 
		FROM nickreg, nickalias WHERE nickreg.id=nickalias.nrid AND nickalias.alias=?");
	$get_greet = $services::dbh->prepare("SELECT nicktext.data FROM nicktext, nickreg, nickid
		WHERE nicktext.nrid=nickid.nrid AND nicktext.type=".NTF_GREET." AND nickid.id=?
		LIMIT 1");
	$del_greet = $services::dbh->prepare("DELETE nicktext.* FROM nicktext, nickreg, nickalias WHERE
		nicktext.type=".NTF_GREET." AND nickreg.id=nickalias.nrid AND nickalias.alias=?");

	$get_num_nicktext_type = $services::dbh->prepare("SELECT COUNT(nicktext.id) FROM nicktext, nickalias
		WHERE nicktext.nrid=nickalias.nrid AND nickalias.alias=? AND nicktext.type=?");
	$drop_nicktext = $services::dbh->prepare("DELETE FROM nicktext USING nickreg, nicktext WHERE
		nicktext.nrid=nickreg.id AND nickreg.nick=?");

	$get_auth_chan = $services::dbh->prepare("SELECT nicktext.data FROM nicktext, nickalias WHERE
		nicktext.nrid=nickalias.nrid AND nicktext.type=(".NTF_AUTH().") AND nickalias.alias=? AND nicktext.chan=?");
	$get_auth_num = $services::dbh->prepare("SELECT nicktext.chan, nicktext.data FROM nicktext, nickalias WHERE 
		nicktext.nrid=nickalias.nrid AND nicktext.type=(".NTF_AUTH().") AND nickalias.alias=? LIMIT 1 OFFSET ?");
	$get_auth_num->bind_param(2, 0, SQL_INTEGER);
	$del_auth = $services::dbh->prepare("DELETE nicktext.* FROM nicktext, nickalias WHERE
		nicktext.nrid=nickalias.nrid AND nicktext.type=(".NTF_AUTH().") AND nickalias.alias=? AND nicktext.chan=?");;
	$list_auth = $services::dbh->prepare("SELECT nicktext.chan, nicktext.data FROM nicktext, nickalias WHERE
		nicktext.nrid=nickalias.nrid AND nicktext.type=(".NTF_AUTH().") AND nickalias.alias=?");

	$set_umode_ntf = $services::dbh->prepare("REPLACE INTO nicktext SELECT nickreg.id, ".NTF_UMODE().", 0, ?, NULL
		FROM nickreg, nickalias WHERE nickreg.id=nickalias.nrid AND nickalias.alias=?");
	$del_umode_ntf = $services::dbh->prepare("DELETE nicktext.* FROM nickreg, nickalias, nicktext
		WHERE nicktext.type=(".NTF_UMODE().") AND nicktext.nrid=nickalias.nrid AND nickalias.alias=?");
	$get_umode_ntf = $services::dbh->prepare("SELECT nicktext.chan FROM nickreg, nickalias, nicktext
		WHERE nicktext.type=(".NTF_UMODE().") AND nicktext.nrid=nickalias.nrid AND nickalias.alias=?");

	$set_vacation_ntf = $services::dbh->prepare("REPLACE INTO nicktext SELECT nickreg.id, ".NTF_VACATION().", 0, ?, NULL
		FROM nickreg, nickalias WHERE nickreg.id=nickalias.nrid AND nickalias.alias=?");
	$get_vacation_ntf = $services::dbh->prepare("SELECT nicktext.chan FROM nickalias, nicktext
		WHERE nicktext.nrid=nickalias.nrid AND nicktext.type=".NTF_VACATION()." AND nickalias.alias=?");

	$set_emailreg_ntf = $services::dbh->prepare("REPLACE INTO nicktext SELECT nickreg.id, ".NTF_EMAILREG().", 0, '', ?
		FROM nickreg, nickalias WHERE nickreg.id=nickalias.nrid AND nickalias.alias=?");
	$get_emailreg_ntf = $services::dbh->prepare("SELECT nicktext.data FROM nickalias, nicktext
		WHERE nicktext.nrid=nickalias.nrid AND nicktext.type=".NTF_EMAILREG()." AND nickalias.alias=?");
	$del_emailreg_ntf = $services::dbh->prepare("DELETE nicktext.* FROM nickreg, nickalias, nicktext
		WHERE nicktext.type=(".NTF_EMAILREG().") AND nicktext.nrid=nickalias.nrid AND nickalias.alias=?");
}


### NICKSERV COMMANDS ###

sub dispatch($$$) {
	my ($src, $dst, $msg) = @_;
	$msg =~ s/^\s+//;
	my @args = split(/\s+/, $msg);
	my $cmd = shift @args;

	my $user = { NICK => $src, AGENT => $nsnick };

	return if operserv::flood_check($user);

	if($cmd =~ /^help$/i) {
		help::sendhelp($nsnick, $src, 'nickserv', @args)
	}
	elsif($cmd =~ /^id(entify)?$/i) {
		if(@args == 1) {
			ns_identify($user, $src, $args[0]);
		} elsif(@args == 2) {
			ns_identify($user, $args[0], $args[1]);
		} else {
			notice($user, 'Syntax: IDENTIFY [nick] <password>');
		}
	}
	elsif($cmd =~ /^sid(entify)?$/i) {
		if(@args == 2) {
			ns_identify($user, $args[0], $args[1], 1);
		} else {
			notice($user, 'Syntax: SIDENTIFY <nick> <password>');
		}
	}
	elsif($cmd =~ /^logout$/i) {
		ns_logout($user);
	}
	elsif($cmd =~ /^release$/i) {
		if(@args == 1) {
			ns_release($user, $args[0]);
		} elsif(@args == 2) {
			ns_release($user, $args[0], $args[1]);
		} else {
			notice($user, 'Syntax: RELEASE <nick> [password]');
		}
	}
	elsif($cmd =~ /^ghost$/i) {
		if(@args == 1) {
			ns_ghost($user, $args[0]);
		} elsif(@args == 2) {
			ns_ghost($user, $args[0], $args[1]);
		} else {
			notice($user, 'Syntax: GHOST <nick> [password]');
		}
	}
	elsif($cmd =~ /^register$/i) {
		if(@args == 2) {
			ns_register($user, $args[0], $args[1]);
		} else {
			notice($user, 'Syntax: REGISTER <password> <email>');
		}
	}
	elsif($cmd =~ /^(?:link|group)$/i) {
		if(@args == 2) {
			ns_link($user, $args[0], $args[1]);
		} else {
			notice($user, 'Syntax: LINK <nick> <password>');
		}
	}
	elsif($cmd =~ /^info$/i) {
		if(@args >= 1) {
			ns_info($user, @args);
		} else {
			notice($user, 'Syntax: INFO <nick> [nick ...]');
		}
	}
	elsif($cmd =~ /^set$/i) {
		ns_set_parse($user, @args);
	}
	elsif($cmd =~ /^(drop|unlink)$/i) {
		if(@args == 1) {
			ns_unlink($user, $src, $args[0]);
		}
		elsif(@args == 2) {
			ns_unlink($user, $args[0], $args[1]);
		}
		else {
			notice($user, 'Syntax: UNLINK [nick] <password>');
		}
	}
	elsif($cmd =~ /^dropgroup$/i) {
		if(@args == 1) {
			ns_dropgroup($user, $src, $args[0]);
		}
		elsif(@args == 2) {
			ns_dropgroup($user, $args[0], $args[1]);
		}
		else {
			notice($user, 'Syntax: DROPGROUP [nick] <password>');
		}
	}
	elsif($cmd =~ /^chgroot$/i) {
		if(@args == 1) {
			ns_changeroot($user, $src, $args[0]);
		}
		elsif(@args == 2) {
			ns_changeroot($user, $args[0], $args[1]);
		}
		else {
			notice($user, 'Syntax: CHGROOT [oldroot] <newroot>');
		}
	}
	elsif($cmd =~ /^sendpass$/i) {
		if(@args == 1) {
			ns_sendpass($user, $args[0]);
		} else {
			notice($user, 'Syntax: SENDPASS <nick>');
		}
	}
	elsif($cmd =~ /^getpass$/i) {
		if(@args == 1) {
			ns_getpass($user, $args[0]);
		} else {
			notice($user, 'Syntax: GETPASS <nick>');
		}
	}
	elsif($cmd =~ /^(?:glist|links)$/i) {
		if(@args == 0) {
			ns_glist($user, $src);
		}
		elsif(@args >= 1) {
			ns_glist($user, @args);
		}
		else {
			notice($user, 'Syntax: GLIST [nick] [nick ...]');
		}
	}
	elsif($cmd =~ /^(?:alist|listchans)$/i) {
		if(@args == 0) {
			ns_alist($user, $src);
		}
		elsif(@args >= 1) {
			ns_alist($user, @args);
		}
		else {
			notice($user, 'Syntax: ALIST [nick] [nick ...]');
		}
	}
	elsif($cmd =~ /^list$/i) {
		if(@args == 1) {
			ns_list($user, $args[0]);
		} else {
			notice($user, 'Syntax: LIST <mask>');
		}
	}
	elsif($cmd =~ /^watch$/i) {
		if ($args[0] =~ /^(add|del|list)$/i) {
			ns_watch($user, $src, @args);
		}
		elsif ($args[1] =~ /^(add|del|list)$/i) {
			ns_watch($user, @args);
		}
		else {
			notice($user, 'Syntax: WATCH <ADD|DEL|LIST> [nick]');
		}
	}
	elsif($cmd =~ /^silence$/i) {
		if ($args[0] =~ /^(add|del|list)$/i) {
			ns_silence($user, $src, @args);
		}
		elsif ($args[1] =~ /^(add|del|list)$/i) {
			ns_silence($user, @args);
		}
		else {
			notice($user, 'Syntax: SILENCE [nick] <ADD|DEL|LIST> [mask] [+expiry] [comment]');
		}
	}
	elsif($cmd =~ /^acc(ess)?$/i) {
		if (@args >= 1) {
			ns_acc($user, @args);
		}
		else {
			notice($user, 'Syntax: ACC <nick>  [nick ...]');
		}
	}
	elsif($cmd =~ /^seen$/i) {
		if(@args >= 1) {
			ns_seen($user, @args);
		}
		else {
			notice($user, 'Syntax: SEEN <nick> [nick ...]');
		}
	}
	elsif($cmd =~ /^recover$/i) {
		if(@args == 1) {
			ns_recover($user, $args[0]);
		} elsif(@args == 2) {
			ns_recover($user, $args[0], $args[1]);
		} else {
			notice($user, 'Syntax: RECOVER <nick> [password]');
		}
	}
	elsif($cmd =~ /^auth$/i) {
		if (@args >= 1) {
			ns_auth($user, @args);
		}
		else {
			notice($user, 'Syntax: AUTH [nick] <LIST|ACCEPT|DECLINE> [num|chan]');
		}
	}
	elsif($cmd =~ /^emailreg$/i) {
		if(@args == 1) {
			ns_emailreg($user, $args[0]);
		} elsif(@args == 2) {
			ns_emailreg($user, $args[0], $args[1]);
		} else {
			notice($user, 'Syntax: EMAILREG [nick] <code>');
		}
	}
	else {
		notice($user, "Unrecognized command.", "For help, type: \002/msg nickserv help\002");
		log::wlog($nsnick, log::DEBUG(), "$src tried to use NickServ $msg");
	}
}

sub ns_identify($$$;$) {
	my ($user, $nick, $pass, $svsnick) = @_;
	my $src = get_user_nick($user);

	my $root = get_root_nick($nick);
	unless($root) {
		notice($user, 'Your nick is not registered.');
		return 0;
	}

	if(is_identified($user, $nick) and !$svsnick) {
		notice($user, 'You are already identified for nick '.$nick.'.');
		return 0;
	}
	elsif($svsnick and lc($src) ne lc($nick) and is_online($nick)) {
		notice($user, $nick.' is already in use. Please use GHOST or RECOVER');
		$svsnick = 0;
	}

	$get_flags->execute($root);
	my ($flags) = $get_flags->fetchrow_array();
	$get_flags->finish();

	if($flags & F_FREEZE) {
		notice($user, "This nick has been frozen and may not be used.", $err_deny);
		services::ulog($nsnick, log::INFO(), "\00305attempted to identify to frozen nick \003\002$nick\002", $user);
		return;
	}

	if($flags & F_EMAILREG) {
		notice($user, "This nick is awaiting an email validation code. Please check your email for instructions.");
		return;
	}

	my $uid = get_user_id($user);
	unless(chk_pass($root, $pass, $user)) {
		if(inc_nick_inval($user)) {
			notice($user, $err_pass);
		}
		services::ulog($nsnick, log::INFO(), "failed to identify to nick $nick (root: $root)", $user);
		return 0;
	}

	$identify_ign->execute($uid, $root);
	$id_update->execute($root, $uid);

	notice($user, 'You are now identified.');

	if($flags & F_VACATION) {
		notice($user, "Welcome back from your vacation, \002$nick\002.");
		$set_vacation_ntf->execute(MIME::Base64::encode(pack('N', time())), $root);
		$set_vacation_ntf->finish();
	}

	my $enforced;
	if(enforcer_quit($nick)) {
		notice($user, 'Your nick has been released from custody.');
		$enforced = 1;
	}

	$get_umode_ntf->execute($nick);
	my ($umodes) = $get_umode_ntf->fetchrow_array();
	$get_umode_ntf->finish();
	if (lc($src) eq lc($nick)) {
		net::setumode($nsnick, $src, '+r'.($umodes ? $umodes : ''));
		$update_nickalias_last->execute($nick); $update_nickalias_last->finish();
	}
	elsif($svsnick) {
		net::svsnick($nsnick, $src, $nick);
		net::setumode($nsnick, $nick, '+r'.($umodes ? $umodes : ''));
		$update_nickalias_last->execute($nick); $update_nickalias_last->finish();
		$user->{NICK} = $nick;
	}
	elsif(defined $umodes) {
		net::setumode($nsnick, $src, $umodes);
	}

	hostserv::hs_on($user, $root, 1);

	nickserv::do_svssilence($user, $root);
	nickserv::do_svswatch($user, $root);

	chanserv::akick_alluser($user);
	chanserv::set_modes_allchan($user) unless $flags & F_NEVEROP;
	
	services::ulog($nsnick, log::INFO(), "identified to nick $nick (root: $root)", $user);

	memoserv::notify($user, $root);
	notify_auths($user, $root) if $flags & F_AUTH;

	return ($enforced ? 2 : 1);
}

sub ns_logout($) {
	my ($user) = @_;
	my $uid = get_user_id($user);
	
	$update_lastseen->execute($uid);
	$logout->execute($uid);
	notice($user, 'You are now logged out');
	services::ulog($nsnick, log::INFO(), "used NickServ LOGOUT", $user);
}

sub ns_release($$;$) {
	my ($user, $nick, $pass) = @_;

	if(chk_flag($nick, F_FREEZE)) {
		notice($user, "This nick has been frozen and may not be used.", $err_deny);
		services::ulog($nsnick, log::INFO(), "\00305attempted to release frozen nick \003\002$nick\002", $user);
		return;
	}

	unless(is_identified($user, $nick)) {
		if($pass) {
			my $s = ns_identify($user, $nick, $pass);
			return if($s == 0); #failed to identify
			if($s == 1) {
				notice($user, "Nick $nick is not being held.");
				return;
			}
		} else {
			notice($user, $err_deny);
			return;
		}
	}

	if(enforcer_quit($nick)) {
		notice($user, 'Your nick has been released from custody.');
	} else {
		notice($user, "Nick $nick is not being held.");
	}
}

sub ns_ghost($$;$) {
	my ($user, $nick, $pass) = @_;
	my $src = get_user_nick($user);

	if(chk_flag($nick, F_FREEZE)) {
		notice($user, "This nick has been frozen and may not be used.", $err_deny);
		services::ulog($nsnick, log::INFO(), "\00305attempted to ghost frozen nick \003\002$nick\002", $user);
		return;
	}

	unless(is_identified($user, $nick)) {
		if($pass) {
			my $s = ns_identify($user, $nick, $pass);
			return if($s == 0); #failed to identify
		} else {
			notice($user, $err_deny);
			return;
		}
	}

	if(!is_online($nick)) {
		notice($user, "\002$nick\002 is not online");
		return;
	} elsif(lc $src eq lc $nick) {
		notice($user, "I'm sorry, $src, I'm afraid I can't do that.");
		return;

	} else {
		net::irckill($nsnick, $nick, "GHOST command used by $src");
		notice($user, "Your ghost has been disconnected");
		services::ulog($nsnick, log::INFO(), "used NickServ GHOST on $nick", $user);
		#nick_delete($nick);
		return;
	}
}

sub ns_register($$$) {
	my ($user, $pass, $email) = @_;
	my $src = get_user_nick($user);
	
	if($src =~ /^guest/i) {
		notice($user, $err_deny);
		return;
	}
	
	unless(validate_email($email)) {
		notice($user, $err_email);
		return;
	}
	
	my $uid = get_user_id($user);
	
	$get_register_lock->execute; $get_register_lock->finish;
	
	if(not is_registered($src)) {
		$register->execute($src, $pass, $email); $register->finish();
		$create_alias->execute($src, $src); $create_alias->finish;
		if (defined($services::conf{'default-protect'})) {
			$set_protect_level->execute((defined($services::conf{'default-protect'}) ?
				$protect_level{lc $services::conf{'default-protect'}} : 1), $src);
			$set_protect_level->finish();
		}
		$unlock_tables->execute; $unlock_tables->finish;
		
		if($services::conf{'validate-email'}) {
			set_flag($src, F_EMAILREG());
			emailreg($src, 1, $email);
			notice($user, "Your registration is not yet complete.", 
				"Your nick will expire within 24 hours if you do not enter the validation code.",
				"Check your email for further instructions.");
		}
		else {
			$identify->execute($uid, $src); $identify->finish();
			notice($user, 'You are now registered and identified.');
			net::setumode($nsnick, $src, '+r');
		}
		
		$id_update->execute($src, $uid); $id_update->finish();
		services::ulog($nsnick, log::INFO(), "registered $src (email: $email)".
			($services::conf{'validate-email'} ? ' requires email validation code' : ''),
			$user);
	} else {
		$unlock_tables->execute; $unlock_tables->finish;
		notice($user, 'Your nickname has already been registered.');
	}
}

sub ns_link($$$) {
	my ($user, $nick, $pass) = @_;

	my $root = get_root_nick($nick);
	my $src = get_user_nick($user);
	my $uid = get_user_id($user);

	if($src =~ /^guest/i) {
		notice($user, $err_deny);
		return;
	}

	unless (is_registered($nick)) {
		if(is_registered($src)) {
			notice($user, "The nick \002$nick\002 is not registered. You need to change your nick to \002$nick\002 and then link to \002$src\002.");
		} else { # if neither $nick nor $src are registered
			notice($user, "You need to register your nick first. For help, type \002/ns help register");
		}
		return;
	}

	unless(chk_pass($root, $pass, $user)) {
		notice($user, $err_pass);
		return;
	}

	if(chk_flag($nick, F_FREEZE) and (lc $pass ne 'force')) {
		notice($user, "\002$root\002 has been frozen and may not be used.");
		return;
	}

	if(is_alias_of($src, $nick)) {
		notice($user, "\002$nick\002 is already linked to \002$src\002.");
		return;
	}

	$get_register_lock->execute; $get_register_lock->finish;
		
	if(is_registered($src)) {
		$unlock_tables->execute; $unlock_tables->finish;
		
		if(is_identified($user, $src)) {
			# This is disabled because it's not used much, it's easy to abuse, and it causes inconsistencies in nickid and chanacc.
			
			notice($user, "You cannot link an already registered nick. Type this and try again: \002/ns drop $src <password>");
			return;
			
			#my $sroot = get_root_nick($src);
			
			## FIXME - This may cause duplicates on access lists
			
			#$change_root->execute($root, $sroot);
			#$change_all_access->execute($root, $sroot);
			#$change_akicks->execute($root, $sroot);
			#$change_founders->execute($root, $sroot);
			#$change_successors->execute($root, $sroot);
			#$change_svsops->execute($root, $sroot);

			#$memoserv::memo_chgroot->execute($root, $sroot);
			#$memoserv::count_ignore->execute($sroot);
			#my $count = $memoserv::count_ignore->fetchrow_array;
			#$memoserv::adjust_ignore->execute($count, $sroot);
			#$memoserv::ignore_chgroot1->execute($root, $sroot);
			#$memoserv::ignore_chgroot2->execute($root, $sroot);

			#$drop->execute($sroot);

			#services::ulog($nsnick, log::INFO(), "merged nicks $sroot and $root.", $user);
		} else {
			notice($user, 'Your nickname has already been registered.');
			return;
		}
	} else {
		$create_alias->execute($src, $root); $create_alias->finish();
		if (defined($services::conf{'default-protect'})) {
			$set_protect_level->execute((defined($services::conf{'default-protect'}) ?
				$protect_level{lc $services::conf{'default-protect'}} : 1), $src);
			$set_protect_level->finish();
		}
		$unlock_tables->execute; $unlock_tables->finish;
		
		if(is_identified($user, $root)) {
			$identify_ign->execute($uid, $root); $identify_ign->finish();
			$id_update->execute($root, $uid); $id_update->finish();
		} else {
			ns_identify($user, $root, $pass);
		}
	}
	
	notice($user, "\002$src\002 is now linked to \002$root\002.");
	services::ulog($nsnick, log::INFO(), "made $src an alias of $root.", $user);

	check_identify($user);
}

sub ns_unlink($$$) {
	my ($user, $nick, $pass) = @_;
	my $uid = get_user_id($user);
	my $src = get_user_nick($user);
	
	my $root = get_root_nick($nick);
	unless(chk_pass($root, $pass, $user)) {
		notice($user, $err_pass);
		return;
	}

	if(chk_flag($nick, F_FREEZE) and (lc $pass ne 'force')) {
		notice($user, "\002$root\002 has been frozen and may not be used.", $err_deny);
		services::ulog($nsnick, log::INFO(), "\00305attempted to unlink \002$nick\002 from frozen nick \002$root\002", $user);
		return;
	}

	if(lc $root eq lc $nick) {
		$count_aliases->execute($root);
		my ($count) = $count_aliases->fetchrow_array;
		if($count == 1) {
			ns_dropgroup_real($user, $root);
			return;
		}

		$get_random_alias->execute($root);
		my ($new) = $get_random_alias->fetchrow_array;
		ns_changeroot($user, $root, $new, 1);
		
		$root = $new;
	}
	
	unidentify_single($nick);
	delete_alias($nick);
	enforcer_quit($nick);
	
	notice($user, "\002$nick\002 has been unlinked from \002$root\002.");
	services::ulog($nsnick, log::INFO(), "removed alias $nick from $root.", $user);
}

sub ns_dropgroup($$$) {
	my ($user, $nick, $pass) = @_;
	my $uid = get_user_id($user);
	my $src = get_user_nick($user);
	my $root = get_root_nick($nick);

	if(adminserv::get_svs_level($root)) {
		notice($user, "A nick with services access may not be dropped.");
		return;
	}

	unless(chk_pass($root, $pass, $user)) {
		notice($user, $err_pass);
		return;
	}

	if(chk_flag($nick, F_FREEZE) and (lc $pass ne 'force')) {
		notice($user, "This nick has been frozen and may not be used.", $err_deny);
		services::ulog($nsnick, log::INFO(), "\00305attempted to dropgroup frozen nick \002$root\002", $user);
		return;
	}

	ns_dropgroup_real($user, $root);
}

sub ns_dropgroup_real($$) {
	my ($user, $root) = @_;
	my $src = get_user_nick($user);
	
	unidentify($root, "Your nick, \002$root\002, was dropped by \002$src\002.", $src);
	dropgroup($root);
	#enforcer_quit($nick);
	notice($user, "Your nick(s) have been dropped.  Thanks for playing.");
	
	services::ulog($nsnick, log::INFO(), "dropped group $root.", $user);
}

sub ns_changeroot($$$;$) {
	my ($user, $old, $new, $force) = @_;

	$force or chk_identified($user, $old) or return;

	my $root = get_root_nick($old);
	
	#unless($oldroot =~ /^\Q$rootnick\E$/i) {
	#	notice($user, "\002$oldroot\002 is not a root nick.  Did you mean \002$rootnick\002?");
	#	return;
	#}

	if(lc($new) eq lc($root)) {
		notice($user, "\002$root\002 is already your root nick.");
		return;
	}
	
	unless(get_root_nick($new) eq $root) {
		notice($user, "\002$new\002 is not an alias of your nick.  Type \002/msg nickserv help link\002 for information about creating aliases.");
		return;
	}

	changeroot($root, $new);

	notice($user, "Your root nick is now \002$new\002.");
	services::ulog($nsnick, log::INFO(), "changed root $root to $new.", $user);
}

sub ns_info($@) {
	my ($user, @nicks) = @_;
	my @reply;

	foreach my $nick (@nicks) {
		my $root = get_root_nick($nick);
	
		$get_info->execute($root);
		my @result = $get_info->fetchrow_array;
		$get_info->finish();

		unless(@result) {
			notice($user, "The nick \002$nick\002 is not registered.");
			next;
		}
	
		my ($email, $regd, $last, $flags, $ident, $vhost, $gecos, $alias_used) = @result;
		# the quit entry might not exist if the user hasn't quit yet.
		$get_nickreg_quit->execute($nick);
		my ($quit) = $get_nickreg_quit->fetchrow_array(); $get_nickreg_quit->finish();
		my $hidemail = $flags & F_HIDEMAIL;

		$get_greet->execute($nick);
		my ($greet) = $get_greet->fetchrow_array(); $get_greet->finish();
		$get_umode_ntf->execute($nick);
		my ($umode) = $get_umode_ntf->fetchrow_array(); $get_umode_ntf->finish();

		my $svslev = adminserv::get_svs_level($root);
		my $protect = protect_level($nick);
		my $showprivate = (is_identified($user, $nick) or
			adminserv::is_svsop($user, adminserv::S_HELP()));
	
		my ($seens, $seenm) = do_seen($nick);

		my @data;
		push @reply, "NickServ info for \002$nick\002:";
		
		push @reply, "  (Online now, $seenm.)" if $seens == 2;
		push @data, ["Last seen:", "$seenm."] if $seens == 1;
		
		push @data,
			["Last seen address:", "$ident\@$vhost"],
			["Registered:", scalar(misc::gmtime2($regd))];
		push @data, ["Last used:", ($last ? misc::gmtime2($last) : 'Unknown')] if $showprivate;
		push @data,
			["Last real name:", $gecos],
			["Last quit:", $quit];
		
		push @data, ["Services Rank:", $adminserv::levels[$svslev]] 
			if $svslev;
		push @data, ["E-mail:", $email] unless $hidemail;
		push @data, ["E-mail:", "$email (Hidden)"] 
			if($hidemail and $showprivate);
		push @data, ["Alias of:", $root] 
			if ((lc $root ne lc $nick) and $showprivate);

		push @reply, fmt::columnar(@data);
		pop @reply; push @reply, ' ';

		push @reply, $protect_long[$protect] if $protect;
		push @reply, "Does not accept memos." if($flags & F_NOMEMO);
		push @reply, "Cannot be added to channel access lists." if($flags & F_NOACC);
		push @reply, "Requires authorization to be added to channel access lists."
			if($flags & F_AUTH);
		push @reply, "Is frozen and may not be used." if($flags & F_FREEZE);
		push @reply, "Will not expire." if($flags & F_HOLD);
		push @reply, "Is currently on vacation." if($flags & F_VACATION);
		push @reply, "Registration pending email-code verification." if($flags & F_EMAILREG);
		push @reply, "UModes on Identify: ".$umode if ($umode and $showprivate);
		push @reply, "Greeting: ".$greet if ($greet);

		pop @reply if($reply[-1] eq ' ');
		push @reply, ' --';
	}

	notice($user, @reply);
}

sub ns_set_parse($@) {
	my ($user, @parms) = @_;
	my $src = get_user_nick($user);
# This is a new NS SET parser
# required due to it's annoying syntax
#
# Most commands have only 2 params at most
# the target (which is implied to be src when not spec'd)
# However in the case of GREET num-params is unbounded
#
# Alternative parsings would be possible,
# one being to use a regexp for valid set/keys
	if (lc($parms[1]) eq 'greet') {
		ns_set($user, @parms);
	}
	elsif(lc($parms[0]) eq 'greet') {
		ns_set($user, $src, @parms);
	}
	else {
		if(@parms == 2) {
			ns_set($user, $src, $parms[0], $parms[1]);
		}
		elsif(@parms == 3) {
			ns_set($user, $parms[0], $parms[1], $parms[2]);
		}
		else {
			notice($user, 'Syntax: SET [nick] <option> <value>');
			return;
		}
	}
}

sub ns_set($$$$) {
	my ($user, $target, $set, @parms) = @_;
	my $src = get_user_nick($user);
	my $override = (adminserv::can_do($user, 'SERVOP') or
		(adminserv::can_do($user, 'FREEZE') and $set =~ /^freeze$/i) ? 1 : 0);
	
	unless(is_registered($target)) {
		notice($user, "$target is not registered.");
		return;
	}
	unless(is_identified($user, $target) or $override) {
		notice($user, $err_deny);
		return;
	}

	unless (
		$set =~ /^protect$/i or
		$set =~ /^e?-?mail$/i or
		$set =~ /^pass(?:w(?:or)?d)?$/i or
		$set =~ /^hidee?-?mail$/i or
		$set =~ /^nomemo$/i or
		$set =~ /^no(?:acc|op)$/i or
		$set =~ /^neverop$/i or
		$set =~ /^auth$/i or
		$set =~ /^(hold|no-?expire)$/i or 
		$set =~ /^freeze$/i or
		$set =~ /^vacation$/i or
		$set =~ /^greet$/i or
		$set =~ /^u?modes?$/i or
		$set =~ /^(email)?reg$/i
	) {
		notice($user, qq{"$set" is not a valid NickServ setting.});
		return;
	}

	my ($subj, $obj);
	if($src eq $target) {
		$subj='Your';
		$obj='You';
	} else {
		$subj="\002$target\002\'s";
		$obj="\002$target\002";
	}

	if($set =~ /^protect$/i) {
		my $level = $protect_level{lc shift @parms};
		unless (defined($level)) {
			notice($user, "Syntax: SET PROTECT <none|normal|high|kill>");
			return;
		}
		
		$set_protect_level->execute($level, $target);
		notice($user, "$subj protection level is now set to \002".$protect_short[$level]."\002. ".$protect_long[$level]);

		return;
	}

	elsif($set =~ /^e?-?mail$/i) {
		unless(@parms == 1) {
			notice($user, 'Syntax: SET EMAIL <address>');
			return;
		}
		my $email = $parms[0];
		
		unless(validate_email($email)) {
			notice($user, $err_email);
			return;
		}

		$set_email->execute($email, $target);
		notice($user, "$subj email address has been changed to \002$email\002.");
		services::ulog($nsnick, log::INFO(), "changed email of \002$target\002 to $email", $user);

		return;
	}

	elsif($set =~ /^pass(?:w(?:or)?d)?$/i) {
		unless(@parms == 1) {
			notice($user, 'Syntax: SET PASSWD <address>');
			return;
		}

		$set_pass->execute($parms[0], $target);
		notice($user, "$subj password has been changed.");
		services::ulog($nsnick, log::INFO(), "changed password of \002$target\002", $user);

		return;
	}

	elsif($set =~ /^greet$/i) {
		unless(@parms) {
			notice($user, 'Syntax: SET [nick] GREET <NONE|greeting>');
			return;
		}

		my $greet = join(' ', @parms);
		if ($greet =~ /^(none|off)$/i) {
			$del_greet->execute($target);
			notice($user, "$subj greet has been deleted.");
			services::ulog($nsnick, log::INFO(), "deleted greet of \002$target\002", $user);
		}
		else {
			$set_greet->execute($greet, $target);
			notice($user, "$subj greet has been set to \002$greet\002");
			services::ulog($nsnick, log::INFO(), "changed greet of \002$target\002", $user);
		}

		return;
	}
	elsif($set =~ /^u?modes?$/i) {
		unless(@parms == 1) {
			notice($user, 'Syntax: SET UMODE <+modes-modes|none>');
			return;
		}

		if (lc $parms[0] eq 'none') {
			$del_umode_ntf->execute($target); $del_umode_ntf->finish();
			notice($user, "umode on identify for \002$obj\002 deleted");
		}
		else {
			my $modes = modes::allowed_umodes($parms[0]);
			$set_umode_ntf->execute($modes, $target); $set_umode_ntf->finish();
			$get_using_nicks->execute($target);
			if($get_using_nicks->rows) {
				while (my ($usernick) = $get_using_nicks->fetchrow_array) {
					net::setumode($nsnick, $usernick, $modes)
				}
			}
			$get_using_nicks->finish();
			notice($user, "umode on identify for \002$obj\002 set to $modes");
		}
		return;
	}

	my $val;
	if($parms[0] =~ /^(?:no|off|false|0)$/i) { $val = 0; }
	elsif($parms[0] =~ /^(?:yes|on|true|1)$/i) { $val = 1; }
	else {
		notice($user, "Please say \002on\002 or \002off\002.");
		return;
	}

	if($set =~ /^hidee?-?mail$/i) {
		set_flag($target, F_HIDEMAIL, $val);
		
		if($val) {
			notice($user, "$subj email address is now hidden.");
		} else {
			notice($user, "$subj email address is now visible.");
		}

		return;
	}

	if($set =~ /^nomemo$/i) {
		set_flag($target, F_NOMEMO, $val);

		if($val) {
			notice($user, "$subj memos will be blocked.");
		} else {
			notice($user, "$subj memos will be delivered.");
		}

		return;
	}
	
	if($set =~ /^no(?:acc|op)$/i) {
		set_flag($target, F_NOACC, $val);

		if($val) {
			notice($user, "$obj may not be added to channel access lists.");
		} else {
			notice($user, "$obj may be added to channel access lists.");
		}

		return;
	}

	if($set =~ /^neverop$/i) {
		set_flag($target, F_NEVEROP, $val);

		if($val) {
			notice($user, "$obj will not be granted status upon joining channels.");
		} else {
			notice($user, "$obj will be granted status upon joining channels.");
		}

		return;
	}

	if($set =~ /^auth$/i) {
		set_flag($target, F_AUTH, $val);

		if($val) {
			notice($user, "$obj must now authorize additions to channel access lists.");
		} else {
			notice($user, "$obj will not be asked to authorize additions to channel access lists.");
		}

		return;
	}

	if($set =~ /^(hold|no-?expire)$/i) {
		unless (adminserv::can_do($user, 'SERVOP') or
			is_identified($user, $target) and adminserv::is_ircop($user))
		{	
			notice($user, $err_deny);
			return;
		}

		set_flag($target, F_HOLD, $val);

		if($val) {
			notice($user, "\002$target\002 is now held from expiration.");
			services::ulog($nsnick, log::INFO(), "has held \002$target\002", $user);
		} else {
			notice($user, "\002$target\002 will now expire normally.");
			services::ulog($nsnick, log::INFO(), "released \002$target\002 from hold", $user);
		}

		return;
	}

	if($set =~ /^freeze$/i) {
		unless (adminserv::can_do($user, 'FREEZE') or
			is_identified($user, $target) and adminserv::is_ircop($user))
		{
			notice($user, $err_deny);
			return;
		}

		set_flag($target, F_FREEZE, $val);

		if($val) {
			notice($user, "\002$target\002 is now frozen.");
			unidentify($target, "Your nick, \002$target\002, has been frozen and may no longer be used.");
			services::ulog($nsnick, log::INFO(), "froze \002$target\002", $user);
		} else {
			notice($user, "\002$target\002 is no longer frozen.");
			services::ulog($nsnick, log::INFO(), "unfroze \002$target\002", $user);
		}

		return;
	}

	if($set =~ /^vacation$/i) {
		if ($val) {
			$get_regd_time->execute($target);
			my ($regd) = $get_regd_time->fetchrow_array;
			$get_regd_time->finish();

			if(($regd > (time() - 86400 * int($services::conf{'vacationexpire'} / 3))) and !$override) {
				notice($user, "$target is not old enough to use VACATION",
					'Minimum age is '.int($services::conf{'vacationexpire'} / 3).' days');
				return;
			}

			$get_vacation_ntf->execute($target);
			my ($last_vacation) = $get_vacation_ntf->fetchrow_array();
			$get_vacation_ntf->finish();
			if(defined($last_vacation)) {
				$last_vacation = unpack('N', MIME::Base64::decode($last_vacation));
				if ($last_vacation > (time() - 86400 * int($services::conf{'vacationexpire'} / 3)) and !$override) {
					notice($user, "I'm sorry, \002$src\002, I'm afraid I can't do that.",
						"Last vacation ended ".misc::gmtime2($last_vacation),
						'Minimum time between vacations is '.int($services::conf{'vacationexpire'} / 3).' days.');
					return;
				}
			}
		}

		set_flag($target, F_VACATION, $val);

		services::ulog($nsnick, log::INFO(),
			($val ? 'enabled' : 'disabled')." vacation mode for \002$target\002", $user);
		notice($user, "Vacation mode ".($val ? 'enabled' : 'disabled')." for \002$target\002");
		return;
	}

	if($set =~ /^(email)?reg$/i) {
		unless (adminserv::can_do($user, 'SERVOP'))
		{
			notice($user, $err_deny);
			return;
		}

		set_flag($target, F_EMAILREG, $val);

		if($val) {
			emailreg($target, 1);
			notice($user, "\002$target\002 now needs an email validation code.");
			unidentify($target, ["Your nick, \002$target\002, has been flagged for an email validation audit.",
				"Your nick will expire within 24 hours if you do not enter the validation code.",
				"Check your email for further instructions."]);
			services::ulog($nsnick, log::INFO(), "requested an email audit for \002$target\002", $user);
		} else {
			$del_emailreg_ntf->execute($target); $del_emailreg_ntf->finish();
			notice($user, "\002$target\002 is now fully registered.");
			services::ulog($nsnick, log::INFO(), "validated the email for \002$target\002", $user);
		}

		return;
	}

}

sub ns_sendpass($$) {
	my ($user, $nick) = @_;

	unless(adminserv::is_svsop($user, adminserv::S_HELP() )) {
		notice($user, $err_deny);
		return;
	}

	$get_pass_email->execute($nick);
	my ($pass, $email) = $get_pass_email->fetchrow_array;

	unless($pass and $email) {
		notice($user, "\002$nick\002 is not registered or does not have an email address.");
		return;
	}

	email::send($email, "$nsnick Password Reminder",
		"The password for the nick $nick is:\n$pass");
	
	notice($user, "Password has been sent to $email.");
	services::ulog($nsnick, log::INFO(), "used SENDPASS on $nick ($email)", $user);
}

sub ns_getpass($$) {
	my ($user, $nick) = @_;

	unless(adminserv::is_svsop($user, adminserv::S_ROOT() )) {
		notice($user, $err_deny);
		return;
	}

	my $pass = get_pass($nick);

	unless($pass) {
		notice($user, "\002$nick\002 is not registered.");
		return;
	}

	notice($user, "Password for \002$nick\002 is: \002$pass\002");
	services::ulog($nsnick, log::INFO(), "used GETPASS on $nick", $user);
}

sub ns_glist($@) {
	my ($user, @targets) = @_;
	my @reply;

	foreach my $target (@targets) {
		my $root = get_root_nick($target);
		unless($root) {
			push @reply, "\002$target\002 is not registered.";
			next;
		}

		unless(is_identified($user, $target) or 
			adminserv::is_svsop($user, adminserv::S_HELP())
		) {
			push @reply, "$target: $err_deny";
			next;
		}

		my @data = (['Nick', 'Protection', 'Last used']);
		$get_glist->execute($root);
		while(my ($alias, $protect, $last) = $get_glist->fetchrow_array) {
			push @data, [$alias, $protect_short[$protect], ($last ? misc::gmtime2($last) : 'Unknown')];
		}

		push @reply,
			"The following ". ($get_glist->rows == 1 ? ' nick is' : $get_glist->rows .' nicks are').
			" in the group of \002$root\002:",
			fmt::columnar(@data);
		$get_glist->finish();
	}

	notice($user, @reply);
}

sub ns_alist($@) {
	my ($user, @targets) = @_;
	my @reply;

	foreach my $target (@targets) {
		(adminserv::is_svsop($user, adminserv::S_HELP()) and (
			chk_registered($user, $target) or next)
		) or chk_identified($user, $target) or next;

		my @data;

		$get_all_access->execute($target);
		while(my ($c, $l, $a, $t) = $get_all_access->fetchrow_array) {
			push @data, [$c, $chanserv::levels[$l], ($a ? "($a)" : ''), ($t ? scalar(misc::gmtime2($t)) : '')];
		}

		push @reply,
			"Access listing for \002$target\002 (".
			$get_all_access->rows." entries)",
			fmt::columnar(@data) ;
	}
	notice($user, @reply);
}

sub ns_list($$) {
	my ($user, $mask) = @_;

	unless(adminserv::is_svsop($user, adminserv::S_HELP())) {
		notice($user, $err_deny);
		return;
	}

	my ($mnick, $mident, $mhost) = misc::parse_mask($mask);
	
	$mnick = '%' if($mnick eq '');
	$mident = '%' if($mident eq '');
	$mhost = '%' if($mhost eq '');

	my @data;
	$get_matching_nicks->execute($mnick, $mident, $mhost);
	while(my ($rnick, $rroot, $rident, $rhost) = $get_matching_nicks->fetchrow_array) {
		push @data, [$rnick, ($rroot ne $rnick ? $rroot : ''), $rident . '@' . $rhost];
	}

	notice($user, 
		"Registered nicks matching \002$mask\002:",
		fmt::columnar(@data)
	);
}

sub ns_watch($$$;$) {
	my ($user, $target, $cmd, $mask) = @_;
	my $src = get_user_nick($user);
	
	my $root = get_root_nick($target);
	unless ($root) {
		notice($user, "\002$target\002 is not registered.");
		return;
	}
	unless(is_identified($user, $target)) {
		notice($user, $err_deny);
		return;
	}
	
	if ($cmd =~ /^add$/i) {
		if($mask =~ /\!/ or $mask =~ /\@/) {
			my ($mnick, $mident, $mhost) = misc::parse_mask($mask);
			if ($mnick =~ /\*/) {
				notice($user, "Invalid mask: \002$mask\002", 
					'A WATCH mask cannot wildcard the nick.');
				return;
			}
		}

		$check_watch->execute($root, $mask);
		if ($check_watch->fetchrow_array) {
			notice($user, "\002$mask\002 is already in \002$target\002's watch list.");
			return;
		}

		$set_watch->execute($mask, time(), $root);
		net::svswatch($nsnick, $src, "+$mask");
		notice($user, "\002$mask\002 added to \002$target\002's watch list.");
		return;
	}
	elsif ($cmd =~ /^del(ete)?$/i) {
		if($mask =~ /\!/ or $mask =~ /\@/) {
			my ($mnick, $mident, $mhost) = misc::parse_mask($mask);
			if ($mnick =~ /\*/) {
				notice($user, "Invalid mask: \002$mask\002", 
					'A WATCH mask cannot wildcard the nick.');
			}
		}

		$check_watch->execute($root, $mask);
		unless ($check_watch->fetchrow_array) {
			notice($user, "\002$mask\002 is not in \002$target\002's watch list.");
			return;
		}
		$del_watch->execute($root, $mask);
		net::svswatch($nsnick, $src, "-$mask");
		notice($user, "\002$mask\002 removed from \002$target\002's watch list.");
	}
	elsif ($cmd =~ /^list$/i) {
		my @data;
		
		$get_watches->execute($root);
		while(my ($mask, $time) = $get_watches->fetchrow_array) {
			push @data, [$mask, scalar(misc::gmtime2($time))];
		}
		
		notice($user,
			"Watch list for \002$target\002:",
			fmt::columnar(@data)
		);
	}
	else {
		notice($user, 'Syntax: WATCH <ADD|DEL|LIST> [nick]');
	}
}

sub ns_silence($$$;$@) {
	my ($user, $target, $cmd, $mask, @args) = @_;
	my ($expiry, $comment);
	my $src = get_user_nick($user);
	
	my $root = get_root_nick($target);
	unless ($root) {
		notice($user, "\002$target\002 is not registered.");
		return;
	}
	
	unless(is_identified($user, $target)) {
		notice($user, $err_deny);
		return;
	}

	if ($cmd =~ /^add$/i) {
		if (substr($args[0],0,1) eq '+') {
			$expiry = shift @args;
		}
		elsif (substr($args[-1],0,1) eq '+') {
			$expiry = pop @args;
		}
		$comment = join(' ', @args);

		unless($mask =~ /\!.+\@/) {
			if ($mask =~ /.+\@.+/) {
				$mask = '*!'.$mask;
			}
			else {
				unless(is_online($mask)) {
					notice($user, "\002$mask\002 is not online or is an invalid mask.");
					return;
				}
				$comment = $mask unless $comment;
				my (undef, $vhost) = get_vhost($mask);
				$mask = '*!*@'.$vhost;
			}
		}

		$check_silence->execute($root, $mask);
		if ($check_silence->fetchrow_array) {
			notice($user, "\002$mask\002 is already in \002$target\002's SILENCE list.");
			return;
		}
		
		if(defined $expiry) {
			$expiry = misc::parse_time($expiry) + time();
		}
		else {
			$expiry = 0;
		};
		$set_silence->execute($mask, time(), $expiry, $comment, $root);
		net::svssilence($nsnick, $src, "+$mask");
		notice($user, "\002$mask\002 added to \002$target\002's SILENCE list.");
	}
	elsif ($cmd =~ /^del(ete)?$/i) {
		unless($mask =~ /\!/ && $mask =~ /\@/) {
			notice($user, "Invalid mask: \002$mask\002", 
				"A SILENCE mask must be in the format nick!ident\@host");
			return;
		}

		$check_silence->execute($root, $mask);
		unless ($check_silence->fetchrow_array) {
			notice($user, "\002$mask\002 is not in \002$target\002's watch list.");
			return;
		}
		$del_silence->execute($root, $mask);
		net::svssilence($nsnick, $src, "-$mask");
		notice($user, "\002$mask\002 removed from \002$target\002's SILENCE list.");
	}
	elsif ($cmd =~ /^list$/i) {
		$get_silences->execute($root);
		
		my @data = ['Mask', 'Comment', 'Time', 'Expiration'];
		while(my ($mask, $time, $expiry, $comment) = $get_silences->fetchrow_array) {
			push @data, [$mask, $comment, scalar(misc::gmtime2($time)),
				($expiry ? "Expires in ".misc::time_rel($expiry-time()) : "Never expires")
			];
		}
		
		notice($user,
			"Silence list for \002$target\002:",
			fmt::columnar(@data)
		);
	}
	else {
		notice($user, 'Syntax: SILENCE [nick] <ADD|DEL|LIST> [mask] [+expiry] [comment]');
	}
}

sub ns_acc($@) {
	my ($user, @targets) = @_;
	my @reply;

	foreach my $target (@targets) {
		unless(is_registered($target)) {
			push @reply, "ACC 0 \002$target\002 is not registered.";
			next;
		}

		unless(is_online($target)) {
			push @reply, "ACC 1 \002$target\002 is registered and offline.";
			next;
		}

		unless(is_identified({NICK => $target}, $target)) {
			push @reply, "ACC 2 \002$target\002 is online but not identified.";
			next;
		}

		push @reply, "ACC 3 \002$target\002 is registered and identified.";
	}
	notice($user, @reply);
}

sub ns_seen($@) {
	my ($user, @nicks) = @_;

	foreach my $nick (@nicks) {
		if(lc $nick eq lc $user->{AGENT}) {
			notice($user, "Oh, a wise guy, eh?");
			next;
		}
		my ($status, $msg) = do_seen($nick);
		if($status == 2) {
			notice($user, "\002$nick\002 is online now, ".$msg.'.');
		} elsif($status == 1) {
			notice($user, "\002$nick\002 was last seen ".$msg.'.');
		} else {
			notice($user, "The nick \002$nick\002 is not registered.");
		}
	}
}

sub ns_recover($$;$) {
	my ($user, $nick, $pass) = @_;
	my $src = get_user_nick($user);

	if(chk_flag($nick, F_FREEZE)) {
		notice($user, "This nick has been frozen and may not be used.", $err_deny);
		services::ulog($nsnick, log::INFO(), "\00305attempted to recover frozen nick \003\002$nick\002", $user);
		return;
	}

	unless(is_identified($user, $nick)) {
		if($pass) {
			my $s = ns_identify($user, $nick, $pass);
			return if($s == 0); #failed to identify
		} else {
			notice($user, $err_deny);
			return;
		}
	}

	if(!is_online($nick)) {
		notice($user, "\002$nick\002 is not online");
		return;
	} elsif(lc $src eq lc $nick) {
		notice($user, "I'm sorry, $src, I'm afraid I can't do that.");
		return;

	} else {
		collide($nick);
		notice($user, "User claiming your nick has been collided", 
			"/msg NickServ RELEASE $nick to get it back before the one-minute timeout.");
		services::ulog($nsnick, log::INFO(), "used NickServ RECOVER on $nick", $user);
		return;
	}
}

sub ns_auth($@) {
	my ($user, @args) = @_;
	my ($target, $cmd);

	if ($args[0] =~ /^(list|accept|approve|decline|reject)$/i) {
		$target = get_user_nick($user);
		$cmd = lc shift @args;
	}
	else {
		$target = shift @args;
		$cmd = lc shift @args;
	}

	unless (is_registered($target)) {
		notice($user, "The nickname \002$target\002 is not registered");
		return;
	}
	unless (is_identified($user, $target)) {
		notice($user, $err_deny);
		return;
	}

	if ($cmd eq 'list') {
		my @reply = ["Channel", "Level", "From", "Time"];
		$list_auth->execute($target);
		while (my ($cn, $data) = $list_auth->fetchrow_array()) {
			my ($adder, $old, $level, $time) = split(':', $data);
			push @reply, [$cn, $chanserv::levels[$level], $adder, misc::gmtime2($time)];
		}
		if ($list_auth->rows()) {
			notice($user, fmt::columnar(@reply));
		}
		else {
			notice($user, "There are no pending authorizations for \002$target\002");
		}
	}
	elsif ($cmd eq 'accept' or $cmd eq 'approve') {
		my $parm = shift @args;
		my ($cn, $adder, $old, $level, $time);
		if(misc::isint($parm) and
			($cn, $adder, $old, $level, $time) = get_auth_num($target, $parm))
		{
		}
		elsif ($parm =~ /^\#/ and 
			($adder, $old, $level, $time) = get_auth_chan($target, $parm))
		{
			$cn = $parm;
		}
		unless ($cn) {
		# This should normally be an 'else' as the elsif above should prove false
		# For some reason, it doesn't work. the unless ($cn) fixes it.
		# It only doesn't work for numbered entries
			notice($user, "There is no entry for \002$parm\002 in \002$target\002's AUTH list");
			return;
		}
		my $chan = { CHAN => $cn };
		my $root = get_root_nick($target);

		# These next 3 lines should use chanserv::set_acc() but it doesn't seem to work.
		# It won't let me use a $nick instead of $user
		$chanserv::set_acc1->execute($cn, $level, $root);
		$chanserv::set_acc2->execute($level, $adder, $cn, $root);
		chanserv::set_modes_allnick($root, $chan, $level) unless chanserv::is_neverop($root);
		
		my $log_str = ($old?'move':'addition')." \002$root\002"
			. ($old ? ' from the '.$chanserv::levels[$old] : '') .
			' to the '.$chanserv::levels[$level]." list of \002$cn\002";
		services::ulog($chanserv::csnick, log::INFO(), "accepted the $log_str from $adder", $user, $chan);
		notice($user, "You have accepted the $log_str");
		$del_auth->execute($target, $cn);
		$del_auth->finish();
		memoserv::send_memo($chanserv::csnick, $adder, "$target accepted the $log_str");
	}
	elsif ($cmd eq 'decline' or $cmd eq 'reject') {
		my $parm = shift @args;
		my ($cn, $adder, $old, $level, $time);
		if(misc::isint($parm) and
			($cn, $adder, $old, $level, $time) = get_auth_num($target, $parm))
		{
		}
		elsif ($parm =~ /^\#/ and 
			($adder, $old, $level, $time) = get_auth_chan($target, $parm))
		{
			$cn = $parm;
		}
		unless ($cn) {
		# This should normally be an 'else' as the elsif above should prove false
		# For some reason, it doesn't work. the unless ($cn) fixes it.
		# It only doesn't work for numbered entries
			notice($user, "There is no entry for \002$parm\002 in \002$target\002's AUTH list");
			return;
		}
		my $chan = { CHAN => $cn };

		my $root = get_root_nick($target);
		my $log_str = ($old?'move':'addition')." \002$root\002"
			. ($old ? ' from the '.$chanserv::levels[$old] : '') .
			' to the '.$chanserv::levels[$level]." list of \002$cn\002";
		services::ulog($chanserv::csnick, log::INFO(), "declined the $log_str from $adder", $user, $chan);
		notice($user, "You have declined $log_str");
		$del_auth->execute($target, $cn);
		$del_auth->finish();
		memoserv::send_memo($chanserv::csnick, $adder, "$target declined the $log_str");
	}
	#elsif ($cmd eq 'read') {
	#}
	else {
		notice($user, "Unknown AUTH cmd");
	}

#These helpers shouldn't be needed anywhere else.
# If they ever are, move them to the helpers section
	sub get_auth_num($$) {
		my ($nick, $num) = @_;
		$get_auth_num->execute($nick, $num - 1);
		my ($cn, $data) = $get_auth_num->fetchrow_array();
		$get_auth_num->finish();
		return ($data ? ($cn, split(/:/, $data)) : undef);
	}
	sub get_auth_chan($$) {
		my ($nick, $cn) = @_;
		$get_auth_chan->execute($nick, $cn);
		my ($data) = $get_auth_chan->fetchrow_array();
		$get_auth_chan->finish();
		return (split(/:/, $data));
	}
}

sub ns_emailreg($$$) {
	my ($user, $target, $code) = @_;

	unless(is_registered($target)) {
		notice($user, "\002$target\002 isn't registered.");
		return;
	}
	unless(chk_flag($target, F_EMAILREG())) {
		notice($user, "\002$target\002 is already validated.");
		return;
	}

	if(emailreg($target, -1, $code)) {
		notice($user, "\002$target\002 is now validated.", "You may now identify to \002$target\002.");
		services::ulog($nsnick, log::INFO(), "validated the email for \002$target\002", $user);
	}
	else {
		notice($user, "\002$target\002 validation failed. Please verify that you typed the code correctly.");
	}
}

### MISCELLANEA ###

sub do_seen($$) {
	my ($nick) = @_;
	my ($status, $msg);
	
	$get_seen->execute($nick);
	if (my ($alias, $root, $lastseen) = $get_seen->fetchrow_array) {
		$nickserv::get_using_nicks->execute($root);
		if($nickserv::get_using_nicks->rows) {
			my @usernicks;
			while (my ($usernick) = $nickserv::get_using_nicks->fetchrow_array) {
				push @usernicks, "\002$usernick\002";
			}
			$status = 2;
			$msg = "using ".(@usernicks==1 ? 'the nick ' : 'the following nicks: ').join(', ', @usernicks);
		}
		else {
			$status = 1;
			$msg = misc::time_ago($lastseen) . " ago (".scalar(misc::gmtime2($lastseen)).")";
		}
	}
	else {
		$status = 0; $msg = undef();
	}

	return ($status, $msg);
}

# For a whole group:
sub unidentify($$;$) {
	my ($nick, $msg, $src) = @_;

	$nick = get_root_nick($nick);

	$get_using_nicks->execute($nick);
	while(my ($t) = $get_using_nicks->fetchrow_array) {
		net::notice($nsnick, $t, (ref $msg ? @$msg : $msg)) unless(lc $t eq lc $src);
		if(is_alias_of($nick, $t)) {
			net::setumode($nsnick, $t, '-r');
		}
	}

	$unidentify->execute($nick);
}

# For a single alias:
sub unidentify_single($$) {
	my ($nick, $msg) = @_;

	if(is_online($nick)) {
		net::setumode($nsnick, $nick, '-r');
	}

	#$unidentify_single->execute($nick);
}

sub notice($@) {
	my $user = shift;
	
	net::notice($user->{AGENT} || $nsnick, get_user_nick($user), @_);
}

sub kill_clones($$) {
	my ($user, $ip) = @_;
	my $uid = get_user_id($user);
	my $src = get_user_nick($user);

	return 0 if $ip == 0;

	$chk_clone_except->execute($uid);
	my ($lim) = $chk_clone_except->fetchrow_array;
	return 0 if $lim == MAX_LIM();
	$lim = $services::conf{'clone-limit'} unless $lim;
	
	$count_clones->execute($ip);
	my ($c) = $count_clones->fetchrow_array;

	if($c > $lim) {
		net::irckill($nsnick, $src, "Session Limit Exceeded");
		return 1;
	}
}

sub kill_user($$) {
	my ($user, $reason) = @_;

	net::irckill(get_user_agent($user) || $config::conf{local}, get_user_nick($user), $reason);
}

sub kline_user($$$) {
	my ($user, $time, $reason) = @_;
	my $agent = get_user_agent($user);
	my ($ident, $host) = nickserv::get_host($user);

	net::kline($agent, '*', $host, $time, $reason);
}

sub validate_email($) {
	my ($email) = @_;

	$email =~ /.+\.(\w+)$/;
	my $tld = $1;
	if(
		$email =~ /^([0-9a-z]+[-._+&])*[0-9a-z]+@([-0-9a-z]+[.])+[a-z]{2,6}$/i and
		$email !~ /^(abuse|postmaster|noc|security|spamtrap)\@/i and
		defined($core::ccode{uc $tld})
	) {
		return 1;
	} else {
		return 0;
	}
}

sub emailreg_code() {
	my $groups = 8; my $length = 5;
	my $emailreg_code;
	for(my $i = 1; $i <= $groups; $i++) {
		for (my $j = 1; $j <= $length; $j++) {
			my $ch;
			$emailreg_code .= (($ch = int(rand(36))) > 9 ? chr(ord('A') + $ch - 10) : $ch);
		}
		$emailreg_code .= '-' unless $i >= $groups;
	}
	return $emailreg_code;
}

sub emailreg($$;$) {
	my ($nick, $sign, $email) = @_;
	if($sign == +1) {
		unless (defined($email)) {
			$get_pass_email->execute($nick);
			(undef, $email) = $get_pass_email->fetchrow_array();
			$get_pass_email->finish();
		}

		my $emailreg_code = emailreg_code();
		$set_emailreg_ntf->execute($emailreg_code, $nick); $set_emailreg_ntf->finish();
		email::send($email, "Nick Registration Code for $nick",
			"Hello $nick,\n\n".
			
			"You are receiving this message from the automated nickname\n".
			"management system of the ".$config::conf{'network'}." network.\n\n".
			
			"If you did not try to register your nickname with us, you can\n".
			"ignore this message. If you continue getting similar e-mails\n".
			"from us, chances are that someone is intentionally abusing your\n".
			"e-mail address. Please contact an administrator for help.\n".
			
			"In order to complete your registration, you must follow the\n".
			"instructions in this e-mail before ".misc::gmtime2(time+86400)."\n".
			
			"To complete the registration, the next time you connect, issue the\n".
			"following command to NickServ:\n\n".

			"/NICKSERV EMAILREG $nick $emailreg_code\n".
			 
			"After you issue the command, your registration will be complete and\n".
			"you will be able to use your nickname.\n\n".
			 
			"---\n".
			"If you feel you have gotten this e-mail in error, please contact\n".
			"an administrator.\n\n".
			 
			"----\n".
			"If this e-mail came to you unsolicited  and appears to be spam -\n".
			"please e-mail ".$config::conf{'replyto'}." with a copy of this e-mail\n".
			"including all headers.\n\n".
			 
			"Thank you.\n");
	}
	elsif($sign == -1) {
		$get_emailreg_ntf->execute($nick); 
		my ($emailreg_code) = $get_emailreg_ntf->fetchrow_array();
		$get_emailreg_ntf->finish();
		if (lc $emailreg_code eq lc $email) {
			set_flag($nick, F_EMAILREG(), 0);
			$del_emailreg_ntf->execute($nick); $del_emailreg_ntf->finish();
			return 1;
		}
		else {
			return 0;
		}
	}
}

# This is mostly for logging, be careful using it for anything else
sub get_hostmask($) {
	my ($user) = @_;
	my ($ident, $host);
	my $src = get_user_nick($user);
	
	($ident, $host) = get_host($user);

	return "$src!$ident\@$host";
}

sub guestnick($) {
	my ($nick) = @_;
	
	$set_guest->execute(1, $nick);
	my $randnick = 'Guest'.int(rand(10)).int(rand(10)).int(rand(10)).int(rand(10)).int(rand(10));
	#Prevent collisions.
	while (is_online($randnick)) {
	    $randnick = 'Guest'.int(rand(10)).int(rand(10)).int(rand(10)).int(rand(10)).int(rand(10));
	}
	net::svsnick($nsnick, $nick, $randnick);

	return $randnick;
}

sub expire {
	my ($ne, $e, $ve, $eve) = ($services::conf{'nearexpire'}, $services::conf{'nickexpire'}, $services::conf{'vacationexpire'}, 
		$services::conf{'validate-expire'});
	
	$get_expired->execute(time() - (86400 * $e), time() - (86400 * $ve), time() - 86400 * $eve);
	while(my ($nick, $email, $ident, $vhost) = $get_expired->fetchrow_array) {
		dropgroup($nick);
		log::wlog($nsnick, log::INFO(), "$nick has expired.  Email: $email  Vhost: $ident\@$vhost");
	}

	my $time = time();

	return unless $ne; # if nearexpire is zero, don't.
	$get_near_expired->execute($time - (86400 * ($e - $ne)), $time - (86400 * ($ve - $ne)));
	while(my ($nick, $email) = $get_near_expired->fetchrow_array) {
		$get_aliases->execute($nick);
		my @aliases;
		while(my ($alias) = $get_aliases->fetchrow_array) {
			push @aliases, $alias;
		}
		
		my $message = "We would like to remind you that your registered nick, $nick, will expire\nin $ne days unless you sign on and identify.";
		if(@aliases > 1) {
			$message .= "\n\nThe following nicks are linked in this group:\n  " . join("\n  ", @aliases);
		}
		
		email::send($email, "$nsnick Expiration Notice", $message);
		
		$set_near_expired->execute($time, 86400 * ($e - ($ne-1)), $time - 86400 * ($e - $ne), $nick);
		log::wlog($nsnick, log::INFO(), "$nick will expire in $ne days. ($email)");
	}
}

sub expire_silence_timed {
	my ($time) = shift;
	$time = 60 unless $time;
	event::addtimer('', $time, __PACKAGE__, 'nickserv::expire_silence_timed');

	expire_silence();
}

sub expire_silence() {
	$get_expired_silences->execute();
	my ($lastnick, @masks);
	while(my ($nick, $mask) = $get_expired_silences->fetchrow_array()) {
		if ($nick eq $lastnick) {
		} else {
			$get_using_nicks->execute($lastnick);
			while(my ($usernick, $userid) = $get_using_nicks->fetchrow_array()) {
				my $user = { NICK => $usernick, ID => $userid, AGENT => $nsnick };
				net::svssilence($nsnick, $usernick, @masks);
				notice($user, "The following SILENCE entries have expired: ".join(', ', @masks));
			}
			$get_using_nicks->finish();
			@masks = ();
			$lastnick = $nick;
		}
		push @masks, '-'.$nick;
	}
	if (@masks) {
		$get_using_nicks->execute($lastnick);
		while(my ($usernick, $userid) = $get_using_nicks->fetchrow_array()) {
			my $user = { NICK => $usernick, ID => $userid, AGENT => $nsnick };
			net::svssilence($nsnick, $usernick, @masks);
			notice($user, "The following SILENCE entries have expired: ".join(', ', @masks));
		}
		$get_using_nicks->finish();
	}
	$get_expired_silences->finish();
	return;
}

sub do_svssilence($$) {
	my ($user, $rootnick) = @_;
	my $target = get_user_nick($user);
	
	$get_silences->execute($rootnick);
	my $count = $get_silences->rows;
	unless ($get_silences->rows) {
		$get_silences->finish;
		return;
	}
	my @silences;
	for(my $i = 1; $i <= $count; $i++) {
		my ($mask, $time, $expiry) = $get_silences->fetchrow_array;
		push @silences, "+$mask";
	}
	$get_silences->finish;
	net::svssilence($nsnick, $target, @silences);
	return;
}

sub do_svswatch($$) {
	my ($user, $rootnick) = @_;
	my $target = get_user_nick($user);
	
	$get_watches->execute($rootnick);
	my $count = $get_watches->rows;
	unless ($get_watches->rows) {
		$get_watches->finish;
		return;
	}
	my @watches;
	for(my $i = 1; $i <= $count; $i++) {
		my ($mask, $time, $expiry) = $get_watches->fetchrow_array;
		push @watches, "+$mask";
	}
	$get_watches->finish;
	net::svswatch($nsnick, $target, @watches);
	return;
}

sub do_umode($$) {
	my ($user, $rootnick) = @_;
	my $target = get_user_nick($user);

	$get_umode_ntf->execute($rootnick);
	my ($umodes) = $get_umode_ntf->fetchrow_array; $get_umode_ntf->finish();

	net::setumode($nsnick, $target, $umodes) if $umodes;
	return
}

sub notify_auths($$) {
	my ($user, $nick) = @_;

	$get_num_nicktext_type->execute($nick, NTF_AUTH);
	my ($count) = $get_num_nicktext_type->fetchrow_array(); $get_num_nicktext_type->finish();
	notice($user, "$nick has $count channel authorizations awaiting action.", 
		"To list them, type /ns auth $nick list") if $count;
}

### PROTECTION AND ENFORCEMENT ###

sub protect($) {
	my ($nick) = @_;

	return if chk_flag($nick, F_FREEZE() | F_EMAILREG());
	my $lev = protect_level($nick);
	
	net::notice($nsnick, $nick,
		"This nickname is registered and protected. If it is your",
		"nick, type \002/msg NickServ IDENTIFY <password>\002. Otherwise,",
		"please choose a different nick."
	) unless($lev==3);

	if($lev == 1) {
		warn_countdown("$nick 60");
	}
	elsif($lev==2) {
		collide($nick);
	}
	elsif($lev==3) {
		net::irckill($nsnick, $nick, "Unauthorized nick use with KILL protection enabled.");
		net::agent_connect($nick, 'enforcer', undef, '+i', 'NickServ Enforcement Agent');
		$enforcers{lc $nick} = 1;
		event::addtimer($nick, 60, __PACKAGE__, "nickserv::enforcer_quit");
	}
	
	return;
}

sub warn_countdown($) {
	my ($cookie)  = @_;
	my ($nick, $rem) = split(/ /, $cookie);
	my $user = { NICK => $nick };
	
	if (is_identified($user, $nick)) {
		$update_nickalias_last->execute($nick); $update_nickalias_last->finish();
		return;
	}
	elsif(!(is_online($nick)) or !(is_registered($nick))) { return; } 

	if($rem == 0) {
		net::notice($nsnick, $nick,
			'Your nick is now being changed.');
		collide($nick);
	} else {
		net::notice($nsnick, $nick,
			"If you do not identify or change your nick in $rem seconds, your nick will be changed.");
		$rem -= 20;
		event::addtimer("$nick $rem", 20, __PACKAGE__, "nickserv::warn_countdown");
	}
}

sub collide($) {
	my ($nick) = @_;
	
	my $newnick = guestnick($nick);
	
	# We have to wait to do the enforcing until the nick actually changes
	$preenforce{lc $nick} = 1;
	event::addtimer($nick, 60, __PACKAGE__, "nickserv::enforcer_quit");
	return $newnick;
}

sub enforcer_quit($) {
	my ($nick) = @_;
	if($enforcers{lc $nick}) {
		undef($enforcers{lc $nick});
		net::agent_quit($nick, 'Enforcer Leaving');
		return 1;
	}
	return 0;
}

### DATABASE UTILITY FUNCTIONS ###

sub get_lock($) {
	my ($nick) = @_;
	
	$nick = lc $nick;

	$user_table++;

	if($cur_lock) {
		if($cur_lock ne $nick) {
			really_release_lock($nick);
			$user_table--;
			die("Tried to get two locks at the same time");
		}
		$cnt_lock++;
	} else {
		$cur_lock = $nick;
		$get_lock->execute($services::conf{'mysql-db'}.".user.$nick");
		$get_lock->finish;
	}
}

sub release_lock($) {
	my ($nick) = @_;
	
	$nick = lc $nick;

	$user_table--;

	if($cur_lock and $cur_lock ne $nick) {
		really_release_lock($cur_lock);
		
		die("Tried to release the wrong lock");
	}
	
	if($cnt_lock) {
		$cnt_lock--;
	} else {
		really_release_lock($nick);
	}
}

sub really_release_lock($) {
	my ($nick) = @_;

	$cnt_lock = 0;
	$release_lock->execute($services::conf{'mysql-db'}.".user.$nick");
	$release_lock->finish;
	undef $cur_lock;
}

sub set_flag($$;$) {
	my ($nick, $flag, $sign) = @_;
	$sign = 1 unless defined($sign);

	if($sign) {
		$set_flag->execute($flag, $flag, $nick);
	} else {
		$unset_flag->execute($flag, $flag, $nick);
	}
}

sub chk_flag($$;$) {
	my ($nick, $flag, $sign) = @_;
	$sign = 1 unless defined($sign);

	$get_flags->execute($nick);
	my ($flags) = $get_flags->fetchrow_array;

	return ($sign ? ($flags & $flag) : !($flags & $flag));
}

sub set_user_flag($$;$) {
	my ($user, $flag, $sign) = @_;
	my $uid = get_user_id($user);
	$sign = 1 unless defined($sign);

	if($sign) {
		$set_flag->execute($flag, $flag, $uid);
	} else {
		$unset_flag->execute($flag, $flag, $uid);
	}
}

sub chk_user_flag($$;$) {
	my ($user, $flag, $sign) = @_;
	my $uid = get_user_id($user);
	$sign = 1 unless defined($sign);

	$get_user_flags->execute($uid);
	my ($flags) = $get_user_flags->fetchrow_array;

	return ($sign ? ($flags & $flag) : !($flags & $flag));
}

sub is_agent($) {
	my ($nick) = @_;
	return (defined($net::agents{lc $nick}));
}

sub get_user_id($) {
	my ($user) = @_;
	my ($id, $n);

	return undef if(is_agent($user->{NICK}) and not $enforcers{lc $user->{NICK}});

	eval { $user->{ID} };
	if($@) {
		die("invalid get_user_id call");
	}

	if(exists($user->{ID})) { return $user->{ID}; }

	my $nick = $user->{NICK};

	# a cheat for isServer()
	if($user->{NICK} =~ /\./) {
		return $user->{ID} = undef;
	}

	if($nick eq '') {
		die("get_user_id called on empty string");
	}

	unless(synced()) {
		lock %uidcache;
		my $x = $uidcache{lc $nick};
		return $x if $x;
	}
	
	while(($n < 10 or (not synced() and !$user_table)) and !defined($id)) {
		$n++;
		$get_user_id->execute($nick);
		($id) = $get_user_id->fetchrow_array;
		unless($id) {
			$get_nickchg->execute($nick);
			($id) = $get_nickchg->fetchrow_array;
		}
	}

	unless($id) { log::wlog($nsnick, log::DEBUG(), "get_user_id($nick) failed. HDLRS: ". join(' ', @net::hdlrs)); }
	
	unless(synced()) {
		lock %uidcache;
		$uidcache{lc $nick} = $id;
	}

	return $user->{ID} = $id;
}


sub get_user_nick($) {
	my ($user) = @_;

	eval { $user->{ID} };
	if($@) {
		die("invalid get_user_nick call");
	}
	
	if(exists($user->{NICK})) { return $user->{NICK} }

	$get_user_nick->execute($user->{ID});
	my ($nick) = $get_user_nick->fetchrow_array;

	return $user->{NICK} = $nick;
}

sub get_user_agent($) {
	my ($user) = @_;

	eval { $user->{AGENT} };
	if($@) {
		die("invalid get_user_agent call");
	}
	
	if(exists($user->{AGENT})) { 
		return $user->{AGENT}
	}
	else {
		return undef;
	}
}

sub get_host($) {
	my ($user) = @_;

	my $id;
	if(ref($user)) {
		$id = get_user_id($user);
	} else {
		$id = get_user_id({ NICK => $user });
	}
	return undef unless $id;

	$get_host->execute($id);
	my ($ident, $host) = $get_host->fetchrow_array;

	return ($ident, $host);
}

sub get_vhost($) {
	my ($user) = @_;

	my $id;
	if(ref($user)) {
		$id = get_user_id($user);
	} else {
		$id = get_user_id({ NICK => $user });
	}
	return undef unless $id;
	
	$get_vhost->execute($id);
	my ($ident, $vhost) = $get_vhost->fetchrow_array;

	return ($ident, $vhost);
}

sub get_user_info($) {
	my ($user) = @_;

	my $uid = get_user_id($user);
	return undef() unless $uid;

	$get_user_info->execute($uid);
	my ($ident, $host, $vhost, $gecos, $server) = $get_user_info->fetchrow_array();
	$get_user_info->finish;

	return ($ident, $host, $vhost, $gecos, $server);
}

sub get_user_modes($) {
	my ($user) = @_;

	my $uid = get_user_id($user);
	$get_umodes->execute($uid);
	my ($umodes) = $get_umodes->fetchrow_array;
	$get_umodes->finish();
	return $umodes;
};

sub set_vhost($$) {
	my ($user, $vhost) = @_;
	my $id = get_user_id($user);
	
	return $set_vhost->execute($vhost, $id);
}

sub set_ident($$) {
	my ($user, $ident) = @_;
	my $id = get_user_id($user);
	
	return $set_ident->execute($ident, $id);
}

sub set_ip($$) {
	my ($user, $ip) = @_;
	my $id = get_user_id($user);

	return $set_ip->execute($ip, $id);
}

sub get_id_nicks($) {
	my ($user) = @_;
	my $id = get_user_id($user);
	my @nicks;

	$get_id_nicks->execute($id);
	while(my ($nick) = $get_id_nicks->fetchrow_array) {
		push @nicks, $nick;
	}

	return (@nicks);
}

sub get_root_nick($) {
	my ($nick) = @_;

	$get_root_nick->execute($nick);
	my ($root) = $get_root_nick->fetchrow_array;

	return $root;
}

sub drop($) {
	my ($nick) = @_;
	
	my $ret = $drop->execute($nick);
	$drop->finish();
	return $ret;
}

sub changeroot($$) {
	my ($old, $new) = @_;

	return if(lc $old eq lc $new);

	$change_root->execute($new, $old);
}

sub dropgroup($) {
	my ($root) = @_;
	
	$del_all_access->execute($root);
	$memoserv::delete_all_memos->execute($root);
	chanserv::drop_nick_chans($root);
	$hostserv::del_vhost->execute($root);
	$drop_watch->execute($root);
	$drop_silence->execute($root);
	$drop_nicktext->execute($root);
	$delete_aliases->execute($root);
	drop($root);
}

sub is_alias($) {
	my ($nick) = @_;
	
	return (get_root_nick($nick) eq $nick);
}

sub delete_alias($) {
	my ($nick) = @_;
	return $delete_alias->execute($nick);
}

sub delete_aliases($) {
	my ($root) = @_;
	return $delete_aliases->execute($root);
}

# How did this get here?
#
#sub get_aliases($) {
#	my ($nick) = @_;
#	
#	$get_aliases->execute($nick);
#	return $get_aliases->fetchrow_array;
#}

sub get_all_access($) {
	my ($nick) = @_;
	
	$get_all_access->execute($nick);
	return $get_all_access->fetchrow_array;
}

sub del_all_access($) {
	my ($root) = @_;
	
	return $del_all_access->execute($root);
}

sub chk_pass($$$) {
	my ($nick, $pass, $user) = @_;

	if(lc($pass) eq 'force' and adminserv::can_do($user, 'SERVOP')) {
		if(adminserv::get_best_svs_level($user) > adminserv::get_svs_level($nick)) {
			return 1;
		}
	}

	$chk_pass->execute($nick, $pass);
	return 1 if($chk_pass->fetchrow_array);
	
	return 0;
}

sub inc_nick_inval($) {
	my ($user) = @_;
	my $id = get_user_id($user);

	$inc_nick_inval->execute($id);
	$get_nick_inval->execute($id);
	my ($nick, $inval) = $get_nick_inval->fetchrow_array;
	if($inval > 3) {
		net::irckill($nsnick, $nick, 'Too many invalid passwords.');
		# unnecessary as irckill calls the quit handler.
		#nick_delete($nick);
		return 0;
	} else {
		return 1;
	}
}

sub is_identified($$) {
	my ($user, $rnick) = @_;
	my $nick = get_user_nick($user);

	$is_identified->execute($nick, $rnick);
	if($is_identified->fetchrow_array) {
		return 1;
	} else {
		return 0;
	}
}

sub chk_identified($;$) {
	my ($user, $nick) = @_;

	$nick = get_user_nick($user) unless $nick;
	
	chk_registered($user, $nick) or return 0;

	unless(is_identified($user, $nick)) {
		notice($user, $err_deny);
		return 0;
	}
	
	return 1;
}

sub is_registered($) {
	my ($nick) = @_;

	$is_registered->execute($nick);
	if($is_registered->fetchrow_array) {
		return 1;
	} else {
		return 0;
	}
}

sub chk_registered($;$) {
	my ($user, $nick) = @_;
	my $src = get_user_nick($user);
	my $what;
	
	if($nick) {
		if(lc $src eq lc $nick) {
			$what = "Your nick";
		} else {
			$what = "The nick \002$nick\002";
		}
	} else {
		$nick = get_user_nick($user) unless $nick;
		$what = "Your nick";
	}

	unless(is_registered($nick)) {
		notice($user, "$what is not registered.");
		return 0;
	}

	return 1;
}

sub is_alias_of($$) {
	$is_alias_of->execute($_[0], $_[1]);
	return ($is_alias_of->fetchrow_array ? 1 : 0);
}

sub check_identify($) {
	my ($user) = @_;
	my $nick = get_user_nick($user);
	if(is_registered($nick)) {
		if(is_identified($user, $nick)) {
			net::setumode($nsnick, $nick, '+r');
			$update_nickalias_last->execute($nick); $update_nickalias_last->finish();
			return 1;
		} else {
			protect($nick);
		}
	}
	return 0;
}

sub protect_level($) {
	my ($nick) = @_;

	$protect_level->execute($nick);
	my ($lev) = $protect_level->fetchrow_array;

	return $lev;
}

sub is_online($) {
	my ($user) = @_;
	my $nick;

	if(ref($user)) {
		$nick = get_user_nick($user);
	} else {
		$nick = $user;
	}
	
	$is_online->execute($nick);
	my ($status) = $is_online->fetchrow_array;
	
	return $status;
}

sub chk_online($$) {
	my ($user, $target) = @_;

	unless(is_online($target)) {
		if(ref($target)) {
			$target = get_user_nick($target);
		}
		
		notice($user, "\002$target\002: No such user.");
		return 0;
	}
	
	return 1;
}

sub get_pass($) {
	my ($nick) = @_;
	
	$get_pass->execute($nick);
	my ($pass) = $get_pass->fetchrow_array;
	
	return $pass;
	
}

sub cleanup() {
	{
		lock %uidcache;
		undef %uidcache;
	}

	while($user_table > 0) { }
	$cleanup->execute();
}

sub cleanup_users() {
	event::addtimer('', OLD_USER_AGE(), __PACKAGE__, 'nickserv::cleanup_users');
	my $time = (time() - OLD_USER_AGE()*2);
	$cleanup_users->execute($time);
	$cleanup_nickid->execute($time);
}

sub fix_vhosts() {
	return; # XXX
	event::addtimer('fix_vhosts', 5, __PACKAGE__, 'nickserv::fix_vhosts');
	$get_hostless_nicks->execute();
	while (my ($nick) = $get_hostless_nicks->fetchrow_array) {
		net::notice($nsnick, $config::conf{'diag'}, "HOSTLESS NICK $nick");
		net::userhost($nick);
		net::userip($nick);
	}
	$get_hostless_nicks->finish();
}

sub nick_cede($) {
	my ($nick) = @_;
	my $id;

	$get_user_id->execute($nick);
	if($id = $get_user_id->fetchrow_array) {
		$nick_id_delete->execute($id);
		$nick_delete->execute($nick);
	}
}

### IRC EVENTS ###

sub nick_create {
	my ($nick, $time, $ident, $host, $vhost, $server, $modes, $gecos, $ip) = @_;
	my $user = { NICK => $nick };
	get_lock($nick);
	
	$nick_check->execute($nick, $time);
	my ($id) = $nick_check->fetchrow_array;

	if($id) {
		$olduser{lc $nick} = 1;
		$nick_create_old->execute($ident, $host, $vhost, $server, $modes, $gecos, UF_FINISHED(), $nick);
	} else {
		nick_cede($nick);
		
		my $flags = (synced() ? UF_FINISHED() : 0);
		my $i;
		while($i < 10 and !$nick_create->execute($nick, $time, $ident, $host, $vhost, $server, $modes, $gecos, $flags)) { $i++ }
	}

	$add_nickchg->execute($net::ircline, $nick, $nick);

	release_lock($nick);

	if($vhost eq '*') {
		net::userhost($nick);
	}
	
	{
		lock %newuser;
		$newuser{lc $nick} = 1;
	}

	if($ip) {
		nickserv::userip(undef, $nick, $ip);
	}
	else {
		net::userip($nick);
	}

	return $id;
}

sub nick_create_post($) {
	my ($nick) = @_;
	my $user = { NICK => $nick };
	my $old = $olduser{lc $nick};
	delete $olduser{lc $nick};

	operserv::do_news($nick, 'u') unless($old);

	get_lock($nick);

	check_identify($user);

	release_lock($nick);
}

sub nick_delete($$) {
	my ($nick, $quit) = @_;
	my $user = { NICK => $nick };
	
	get_lock($nick);
	
	my $id = get_user_id($user);

	$del_nickchg_id->execute($id); $del_nickchg_id->finish();

	$quit_update->execute($quit, $id); $quit_update->finish();
	$update_lastseen->execute($id); $update_lastseen->finish();

	$get_quit_empty_chans->execute($id);

	$nick_chan_delete->execute($id); $nick_chan_delete->finish();
	$nick_quit->execute($nick); $nick_quit->finish();

	release_lock($nick);

	while(my ($cn) = $get_quit_empty_chans->fetchrow_array) {
		chanserv::channel_emptied({CHAN => $cn});
	}
	$get_quit_empty_chans->finish();
}

sub squit($$$) {
	my (undef, $server, $reason) = @_;

	$get_squit_lock->execute; $get_squit_lock->finish;

	$get_squit_empty_chans->execute($server);
	
	$squit_nickreg->execute("Netsplit from $server", $server);
	$squit_nickreg->finish;

	$squit_users->execute($server);
	$squit_users->finish;

	while(my ($cn) = $get_squit_empty_chans->fetchrow_array) {
		chanserv::channel_emptied({CHAN => $cn});
	}
	$get_squit_empty_chans->finish;

	$unlock_tables->execute; $unlock_tables->finish;
}

sub nick_change($$$) {
	my ($old, $new, $time) = @_;

	return if(lc $old eq lc $new);

	if($preenforce{lc $old}) {
		net::agent_connect($old, 'enforcer', undef, '+i', 'NickServ Enforcement Agent');
		undef($preenforce{lc $old});
		$enforcers{lc $old} = 1;
		event::addtimer($old, 60, __PACKAGE__, "nickserv::enforcer_quit");
	}

	$user_table++;
	get_lock($old);
	nick_cede($new);
	$nick_change->execute($new, $time, $old);
	$add_nickchg->execute($net::ircline, $new, $new);
	$user_table--;
	release_lock($old);

	if($new =~ /^guest/i) {
		$get_guest->execute($new);
		if($get_guest->fetchrow_array) {
			$set_guest->execute(0, $new);
		} else {
			guestnick($new);
		}
		return;
	}
	
	net::setumode($nsnick, $new, '-r') 
		unless check_identify({ NICK => $new });
}

sub umode($$) {
	my ($nick, $modes) = @_;
	my $user = { NICK => $nick };

	get_lock($nick);

	my $id = get_user_id($user);
	
	$get_umodes->execute($id);
	my ($omodes) = $get_umodes->fetchrow_array;
	$set_umodes->execute(modes::add($omodes, $modes, 0), $id);

	release_lock($nick);

	my %modelist = modes::splitumodes($modes);
	if (defined($modelist{x}) and $modelist{x} eq '-') {
		my ($ident, $host) = get_host($user);
		chghost(undef, $nick, $host)
	}
	elsif((defined($modelist{x}) and $modelist{x} eq '+') and !defined($modelist{t}) ) {
		net::userhost($nick);
	}
	# Else we will get it in a sethost or chghost
	# Also be aware, our tracking of umodes xt is imperfect
	# as the ircd doesn't always report it to us
	# This might need fixing up in chghost()
}

sub killhandle($$$$) {
	my ($src, $dst, $path, $reason) = @_;
	unless (exists($net::agents{$dst})) {
		nick_delete($dst, "Killed ($src ($reason))");
	}
}

sub userip($$$) {
	my($src, $nick, $ip) = @_;
	my $user = { 'NICK' => $nick };
	my $new = $newuser{lc $nick};
	delete $newuser{lc $nick};
	#my $targetid = get_nick_id($target);
	my $iip; my @ips = split(/\./, $ip);
	for(my $i; $i < 4; $i++) {
		$iip += $ips[$i] * (2 ** ((3 - $i) * 8));
	}

	get_lock($nick);
	
	my $id = get_user_id($user);
	set_ip($user, $iip);
	my $killed = kill_clones($user, $iip);

	release_lock($nick);

	nick_create_post($nick) if(!$killed and $new);
}

sub chghost($$$) {
	my ($src, $dst, $vhost) = @_;
	my $user = { NICK => $dst };
	my $uid = get_user_id($user);
	
	#my ($cur_ident, $cur_vhost) = get_vhost($targetid);
	#if $cur_vhost eq '*' {
	$update_regnick_vhost->execute($vhost, $uid);
	$update_regnick_vhost->finish();
	#}
	
	get_lock($dst);
	
	set_vhost($user, $vhost);
	chanserv::akick_alluser($user);

	release_lock($dst);
}

sub chgident($$$) {
	my ($src, $dst, $ident) = @_;
	my $user = { NICK => $dst };
	
	set_ident($user, $ident);
	chanserv::akick_alluser($user);
}

1;
